Package: PatientGenerator
Type: Package
Title: Generator of Synthetic Patient Data for the OMOP Common Data
        Model
Version: 0.1.4
Authors@R: c(
    person("Cesar", "Barboza", , "c.barboza@mi-erasmusmc.nl", role = c("aut", "cre"), comment = c(ORCID = "0009-0002-4453-3071")),
    person("Ger", "Inberg", , "g.inberg@erasmusmc.nl", role = "aut", comment = c(ORCID = "0000-0001-8993-8748")),
    person("Adam", "Black", , "black@ohdsi.org", role = "aut", comment = c(ORCID = "0000-0001-5576-8701"))
  )
Description: Tools to generate synthetic patient-level test datasets in the
    Observational Medical Outcomes Partnership (OMOP) Common Data Model (CDM).
    Includes a chat-driven generator backed by large language models and an
    interactive 'shiny' designer for editing CDM test sets.
URL: https://github.com/mi-erasmusmc/PatientGenerator,
        https://mi-erasmusmc.github.io/PatientGenerator/
BugReports: https://github.com/mi-erasmusmc/PatientGenerator/issues
License: Apache License (>= 2)
Encoding: UTF-8
RoxygenNote: 7.3.3
Depends: R (>= 4.1.0)
Imports: bslib, checkmate, data.table, DBI, dplyr, DT, duckdb, ellmer,
        glue, httr2, jsonlite, r2d3, R6, shiny, stringr, testthat
Suggests: CDMConnector, CohortCharacteristics, CohortConstructor, covr,
        ggplot2, knitr, rmarkdown, TestGenerator
VignetteBuilder: knitr
Config/Needs/website: pkgdown
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-04-27 09:48:57 UTC; cbarboza
Author: Cesar Barboza [aut, cre] (ORCID:
    <https://orcid.org/0009-0002-4453-3071>),
  Ger Inberg [aut] (ORCID: <https://orcid.org/0000-0001-8993-8748>),
  Adam Black [aut] (ORCID: <https://orcid.org/0000-0001-5576-8701>)
Maintainer: Cesar Barboza <c.barboza@mi-erasmusmc.nl>
Repository: CRAN
Date/Publication: 2026-05-04 19:10:02 UTC
Built: R 4.5.2; ; 2026-05-04 23:14:20 UTC; unix

# Generated by roxygen2: do not edit by hand

export("%||%")
export(availableModels)
export(conceptSearchServer)
export(conceptSearchUI)
export(getTestSets)
export(hecateClient)
export(hecateSearch)
export(patientChat)
export(patientChatNaive)
export(patientDesigner)
import(bslib)
import(dplyr)
import(r2d3)
import(shiny)
importFrom(R6,R6Class)
importFrom(checkmate,assertCharacter)
importFrom(checkmate,assertDirectoryExists)
importFrom(checkmate,assertFileExists)
importFrom(data.table,as.data.table)
importFrom(data.table,rbindlist)
importFrom(data.table,set)
importFrom(ellmer,chat_openai)
importFrom(ellmer,type_from_schema)
importFrom(httr2,req_headers)
importFrom(httr2,req_perform)
importFrom(httr2,request)
importFrom(httr2,resp_body_json)
importFrom(jsonlite,fromJSON)
importFrom(stats,setNames)
importFrom(stringr,str_remove)
importFrom(tools,file_path_sans_ext)
importFrom(utils,tail)

# PatientGenerator

<!-- badges: start -->

[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
[![R-CMD-check](https://github.com/mi-erasmusmc/PatientGenerator/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/mi-erasmusmc/PatientGenerator/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/mi-erasmusmc/PatientGenerator/graph/badge.svg)](https://app.codecov.io/gh/mi-erasmusmc/PatientGenerator)
<!-- badges: end -->

`PatientGenerator` facilitates the creation of synthetic test datasets
for the OMOP Common Data Model (CDM) using two complementary approaches:

- **`patientChat`**: Generates structured patient JSON files using Large
  Language Models (LLMs).
- **`patientDesigner`**: Provides a D3-based Shiny interface for
  reviewing and editing CDM test sets.

The package also includes support for Hecate-powered concept lookups to
ensure valid OMOP concept codes.

### Installation

``` r
# install.packages("remotes")
remotes::install_github("mi-erasmusmc/PatientGenerator")
```

### Workflow Overview

1.  **Generate** an initial synthetic cohort using `patientChat`.
2.  **Save** JSON test sets to the local filesystem.
3.  **Refine** patients using `patientDesigner()`.
    - Utilize built-in concept search (powered by `hecateSearch`) during
      table editing.

### Synthetic Patient Generation with `patientChat`

Set an `OPENAI_API_KEY` environment variable (e.g., via
`usethis::edit_r_environ()`) to enable LLM access.

Available models can be listed using
`PatientGenerator::availableModels()`.

``` r
library(PatientGenerator)

patientGenerator <- patientChat$new(
  model = "gpt-5.4",
  echo = "none"
)
```

### Generating Patients via Natural Language Prompts

Provide detailed prompts, including specific concept sets, for optimal
results.

``` r
patientGenerator$prompt(
  "Population (person table):
     - 10 adult patients
     - 5 female
     - 5 male
  
   Observation Period:
     - Start date between date of birth and 2025-12-31
  
   Condition Occurrence:
     - All patients must have Diabetes (condition_concept_id: 201826)
     - Start date between 2015-01-01 and 2020-12-31
  
   Drug Exposure:
     - All patients must have Semaglutide (drug_concept_id: 19079450)
     - Exposure within 30 days post-index date
  
   Measurement:
     - All patients must have Fasting glucose (measurement_concept_id: 3018251)
  
   Procedure Occurrence:
     - 50% of patients must have Amputation of toe (procedure_concept_id: 4159766)
  
   Output Requirements:
     - Populate only the tables specified in this prompt"
)
```

### Integration with `testthat`

Save the generated dataset as a JSON file and utilize
`TestGenerator::patientsCDM` to instantiate a CDM reference.

``` r
patientGenerator$save(name = "diabetes-patients")

cdm <- TestGenerator::patientsCDM(
  testName = "diabetes-patients",
  cdmVersion = "5.4"
)

cdm$person |> 
  collect() |> 
  print()
```

    #> cdm$person |> collect() |> head(5)
    #>    person_id gender_concept_id year_of_birth person_source_value
    #>        <int>             <int>         <int>              <char>
    #> 1:         1              8532          1965              SYN001
    #> 2:         2              8532          1972              SYN002
    #> 3:         3              8532          1958              SYN003
    #> 4:         4              8532          1981              SYN004
    #> 5:         5              8532          1949              SYN005

### Iterative Refinement

The LLM can be instructed to modify the current test set within the same
`patientChat` instance.

``` r
patientGenerator$prompt("Remove all male patients")
```

    #> cdm$person |> collect() |> head(5)
    #>    person_id gender_concept_id year_of_birth person_source_value
    #>        <int>             <int>         <int>              <char>
    #> 1:         1              8532          1965              SYN001
    #> 2:         2              8532          1972              SYN002
    #> 3:         3              8532          1958              SYN003
    #> 4:         4              8532          1981              SYN004
    #> 5:         5              8532          1949              SYN005

### Visual Review and Editing with `patientDesigner()`

Launch the interactive editor to review and refine datasets:

``` r
PatientGenerator::patientDesigner()
```

The interface supports:

- Loading existing JSON test sets.
- Interactive CRUD operations (Create, Read, Update, Delete) on CDM
  tables.
- Visual timeline inspection and table previews.
- Exporting updated test sets to JSON.

### Concept Search with Hecate

`patientDesigner` integrates a concept search module powered by
`hecateSearch()`. This allows users to search for and insert valid OMOP
concept IDs directly into the CDM tables.

Configure Hecate globally via environment variables:

``` r
Sys.setenv(
  HECATE_BASE_URL = "https://your-hecate-server/api",
  HECATE_API_KEY = "your-api-key"
)
```

Or via package options:

``` r
options(PatientGenerator.hecate = list(
  base_url = "https://your-hecate-server/api",
  timeout_ms = 15000,
  api_key = "your-api-key"
))
```

### Further Documentation

- **Vignette**:
  `vignette("shiny-integration", package = "PatientGenerator")`
- **Reference**: Detailed API documentation and benchmarks are available
  on the GitHub Pages site.

# Copyright 2023 DARWIN EU®
#
# Hecate vocabulary search API client and search helpers.
# Uses camelCase and namespaced calls (httr2::, dplyr::).

#' Get Hecate client configuration
#'
#' Reads configuration from \code{options("patientgenerator.hecate")} (list with
#' \code{base_url}, \code{timeout_ms}, \code{api_key}) or environment variables
#' \code{HECATE_BASE_URL}, \code{HECATE_TIMEOUT_MS}, \code{HECATE_API_KEY}.
#' If the default host is unreachable (e.g. "Could not resolve host"),
#' set \code{Sys.setenv(HECATE_BASE_URL = "https://your-hecate-server")} to
#' point to your Hecate instance.
#'
#' @return A list with \code{baseUrl}, \code{timeoutMs}, \code{apiKey}.
#' @noRd
getHecateConfig <- function() {
  cfg <- getOption("patientgenerator.hecate", list())
  list(
    baseUrl = cfg$base_url %||% Sys.getenv("HECATE_BASE_URL", "https://hecate.pantheon-hds.com/api"),
    timeoutMs = cfg$timeout_ms %||% as.integer(Sys.getenv("HECATE_TIMEOUT_MS", "10000")),
    apiKey = cfg$api_key %||% Sys.getenv("HECATE_API_KEY", "")
  )
}

#' Null coalescing operator
#'
#' @param x First value (any type).
#' @param y Fallback value when \code{x} is NULL.
#' @return \code{x} if not NULL, otherwise \code{y}.
#' @name nullOr
#' @export
`%||%` <- function(x, y) if (is.null(x)) y else x

#' Create a rate limiter
#'
#' @param maxCalls Maximum calls allowed in the time window.
#' @param perSeconds Time window in seconds.
#' @return A function that blocks when the rate limit would be exceeded.
#' @noRd
createRateLimiter <- function(maxCalls = 100, perSeconds = 60) {
  callTimes <- numeric(0)
  function() {
    now <- as.numeric(Sys.time())
    callTimes <<- callTimes[callTimes > (now - perSeconds)]
    if (length(callTimes) >= maxCalls) {
      waitTime <- perSeconds - (now - min(callTimes))
      Sys.sleep(max(waitTime, 0))
      callTimes <<- numeric(0)
    }
    callTimes <<- c(callTimes, now)
    invisible(NULL)
  }
}

hecateRateLimiter <- createRateLimiter(100, 60)

#' Create a Hecate API client
#'
#' @param baseUrl Base URL of the Hecate API (default from config).
#' @param timeoutMs Timeout in milliseconds (default from config).
#' @param apiKey Optional API key for authorization (default from config).
#' @return A client object with class \code{hecate_client}.
#' @export
hecateClient <- function(baseUrl = NULL, timeoutMs = NULL, apiKey = NULL) {
  cfg <- getHecateConfig()
  baseUrl <- baseUrl %||% cfg$baseUrl
  timeoutMs <- timeoutMs %||% cfg$timeoutMs
  apiKey <- apiKey %||% cfg$apiKey

  if (nzchar(apiKey) && nchar(apiKey) < 10) {
    warning("API key appears too short. Please check HECATE_API_KEY environment variable.", call. = FALSE)
  }

  if (isTRUE(getOption("app.debug"))) {
    message("Initializing Hecate client with URL: ", baseUrl)
    message("API key ", if (nzchar(apiKey)) "provided" else "not provided")
  }

  structure(
    list(
      baseUrl = sub("/+$", "", baseUrl),
      timeoutMs = timeoutMs,
      apiKey = apiKey
    ),
    class = "hecate_client"
  )
}

#' Build a Hecate API request
#' @noRd
hecateRequest <- function(client, path, query = NULL) {
  url <- paste0(client$baseUrl, "/", sub("^/+", "", path))
  req <- httr2::request(url) |>
    httr2::req_timeout(client$timeoutMs / 1000)

  if (nzchar(client$apiKey)) {
    req <- httr2::req_headers(req, Authorization = paste("Bearer", client$apiKey))
  }

  if (!is.null(query)) {
    query <- Filter(Negate(is.null), query)
    req <- httr2::req_url_query(req, !!!query)
  }

  req
}

#' Perform a Hecate API request and return parsed JSON or error list
#' Handles connection failures (e.g. unresolved host) and HTTP errors.
#' @noRd
hecatePerform <- function(req) {
  hecateRateLimiter()
  resp <- tryCatch(httr2::req_perform(req), error = function(e) e)

  if (inherits(resp, "error")) {
    msg <- conditionMessage(resp)
    code <- "request_failed"
    if (grepl("Could not resolve host|resolve hostname|getaddrinfo", msg, ignore.case = TRUE)) {
      code <- "host_unreachable"
    }
    return(list(error = "Request failed", code = code, message = msg))
  }

  bodyTxt <- httr2::resp_body_string(resp)
  parsed <- tryCatch(jsonlite::fromJSON(bodyTxt, simplifyVector = FALSE), error = function(e) NULL)

  if (is.null(parsed)) {
    return(list(error = "Invalid JSON response", code = "parse_error", message = bodyTxt))
  }
  parsed
}

#' Search Hecate concepts and return results as a data frame
#'
#' @param query Character(1); search query (required).
#' @param vocabularyId Character(1) or NULL; optional vocabulary filter (comma-separated).
#' @param standardConcept Character(1) or NULL; e.g. \code{"S"} (Standard), \code{"C"} (Classification).
#' @param domainId Character(1) or NULL; optional domain filter (comma-separated).
#' @param conceptClassId Character(1) or NULL; optional concept class filter.
#' @param limit Integer(1); max results (default 20, max 150).
#' @param client Hecate client (default \code{hecateClient()}).
#' @return Data frame of search results, or NULL if an error occurred (API error or bad response shape).
#'
#' @examples
#' \dontrun{
#' # Simple search
#' df <- hecateSearch("diabetes")
#'
#' # Search with filters
#' df <- hecateSearch("hypertension", domainId = "Condition", limit = 10)
#' }
#' @export
hecateSearch <- function(
    query,
    vocabularyId = NULL,
    standardConcept = NULL,
    domainId = NULL,
    conceptClassId = NULL,
    limit = 20,
    client = hecateClient()) {
  if (!is.character(query) || length(query) != 1 || is.na(query) || nchar(query) < 1) {
    stop("`query` must be a non-empty string.")
  }

  req <- hecateRequest(
    client,
    path = "search",
    query = list(
      q = query,
      vocabulary_id = vocabularyId,
      standard_concept = standardConcept,
      domain_id = domainId,
      concept_class_id = conceptClassId,
      limit = limit
    )
  )

  res <- hecatePerform(req)

  if (is.list(res) && "error" %in% names(res)) {
    detailStr <- if (!is.null(res$message)) paste(" -", res$message) else ""
    if (identical(res$code, "host_unreachable")) {
      warning(
        "Hecate API unreachable: ", res$error, detailStr, "\n",
        "  Set HECATE_BASE_URL to your Hecate instance (e.g. Sys.setenv(HECATE_BASE_URL = \"https://your-hecate-server\")).",
        call. = FALSE
      )
    } else {
      warning("API error: ", res$error, detailStr, call. = FALSE)
    }
    return(NULL)
  }

  parsed <- res

  if (!is.list(parsed) || length(parsed) == 0) {
    warning("Unexpected response format or empty response")
    return(NULL)
  }

  if (length(parsed) > 0) {
    firstElem <- parsed[[1]]
    if (!is.list(firstElem) || !"concepts" %in% names(firstElem)) {
      warning("Response does not match expected structure (missing 'concepts' field)")
      return(NULL)
    }
  }

  resultRows <- list()

  for (i in seq_along(parsed)) {
    searchResult <- parsed[[i]]

    searchConceptName <- if (is.null(searchResult$concept_name)) NA_character_ else searchResult$concept_name
    searchConceptNameLower <- if (is.null(searchResult$concept_name_lower)) NA_character_ else searchResult$concept_name_lower
    searchScore <- if (is.null(searchResult$score)) NA_real_ else searchResult$score

    conceptsList <- searchResult$concepts
    if (is.null(conceptsList) || length(conceptsList) == 0) {
      next
    }

    for (j in seq_along(conceptsList)) {
      concept <- conceptsList[[j]]

      row <- list(
        searchConceptName = searchConceptName,
        searchConceptNameLower = searchConceptNameLower,
        searchScore = searchScore,
        conceptId = if (is.null(concept$concept_id)) NA_integer_ else concept$concept_id,
        conceptName = if (is.null(concept$concept_name)) NA_character_ else concept$concept_name,
        domainId = if (is.null(concept$domain_id)) NA_character_ else concept$domain_id,
        vocabularyId = if (is.null(concept$vocabulary_id)) NA_character_ else concept$vocabulary_id,
        conceptClassId = if (is.null(concept$concept_class_id)) NA_character_ else concept$concept_class_id,
        standardConcept = if (is.null(concept$standard_concept)) NA_character_ else concept$standard_concept,
        conceptCode = if (is.null(concept$concept_code)) NA_character_ else concept$concept_code,
        invalidReason = if (is.null(concept$invalid_reason)) NA_character_ else concept$invalid_reason,
        validStartDate = if (is.null(concept$valid_start_date)) NA_character_ else concept$valid_start_date,
        validEndDate = if (is.null(concept$valid_end_date)) NA_character_ else concept$valid_end_date,
        recordCount = if (is.null(concept$record_count)) NA_integer_ else concept$record_count
      )

      resultRows <- c(resultRows, list(row))
    }
  }

  if (length(resultRows) == 0) {
    warning("No concepts found in response")
    return(NULL)
  }

  tryCatch(
    {
      df <- dplyr::bind_rows(resultRows)
      return(df)
    },
    error = function(e) {
      warning("Error converting to dataframe: ", conditionMessage(e))
      return(NULL)
    }
  )
}

#' `patientChatNaive()` is a grapper for the ellmer package to send prompts and send the output test set to an LLM. Requires a valid API key.
#'
#' @description
#'    Priorities:
#'      - Accepts a prompt as an input.
#'      - Produces a test set in accordance to the provided JSON schema.
#'      - Utilizes tools such as CodelistGenerator or Hecate to look up for functions.
#'      - Accepts a subsequent prompt with a test set that the LLM has to use as a context
#'
#'    One function for this tasks allow us to:
#'      - Test the test sets created by the LLM.
#'      - Test prompt engineering.
#'      - Test integration of tools functionality.
#'      - Allow us to create fast a small set of patients to test analytical packages.
#'
#' @param prompt A prompt to the LLM, in character or JSON response.
#' @param model The model used by the LLM. Currently only OpenAI models are accepted.
#' @param jsonSchemaPath Path to a JSON schema used to structure the response.
#'
#' @returns A JSON response that includes: the natural language answer from the LLM and a JSON with test set patients in accordance to the provided schema.
#' @importFrom ellmer chat_openai
#' @importFrom ellmer type_from_schema
#' @importFrom jsonlite fromJSON
#' @export
patientChatNaive <- function(prompt = "### Give me a sample of five patients",
                             model = "gpt-5.2",
                             jsonSchemaPath = NULL) {

  # Check params ---------------------------------------------------------------
  checkmate::assertCharacter(prompt)
  checkmate::assertCharacter(model)
  if (!is.null(jsonSchemaPath)) {
    # browser()
    checkmate::assertCharacter(jsonSchemaPath)
    tryCatch({
      jsonlite::fromJSON(jsonSchemaPath)
    }, error = function(e) {
      stop("jsonSchemaPath doesn't lead to a JSON file")
    })
  } else {
    jsonSchemaPath <- system.file("jsonSchemas",
                                  "cdm54schema-complete.json",
                                  package = "PatientGenerator")
    checkmate::assertFileExists(jsonSchemaPath)
  }

  # Check API and available models ---------------------------------------------
  api_models <- availableModels()
  if (!model %in% api_models) {
    stop(
      glue::glue("{model} not available.\n"),
      "\n These are some models that are available to you:\n",
      paste0("  - ", sample(api_models, 10), collapse = "\n"),
      "\n For a complete list, call availableModels()\n",
      call. = FALSE
    )
  }

  # Chat -----------------------------------------------------------------------
  chat <- ellmer::chat_openai(model = model)
  response <- chat$chat_structured(
    prompt,
    type = ellmer::type_from_schema(
      path = jsonSchemaPath
    )
  )
  return(response)
}

#' `availableModels()` If the API key is valid in the system, returns available models to the user from the LLM provider.
#'
#' @description OpenAI is the only one currently supported.
#'
#' @returns A string list with the id of a vailable models.
#' @importFrom httr2 request req_headers req_perform resp_body_json
#' @export
availableModels <- function() {
  check_api_key()
  # Retrieve and check available models
  response <- httr2::request("https://api.openai.com/v1/models") |>
    httr2::req_headers(
      Authorization = paste("Bearer", Sys.getenv("OPENAI_API_KEY"))
    ) |>
    httr2::req_perform()
  models <- resp_body_json(response)
  lapply(models$data, function(data) data$id) |>
    unlist()
}

check_api_key <- function(apiKeyName = "OPENAI_API_KEY") {
  key <- Sys.getenv("OPENAI_API_KEY", unset = "")
  if (!nzchar(key)) {
    stop(
      "API key not found.\n",
      "Set it with Sys.setenv(OPENAI_API_KEY = 'your_key') ",
      "and/or add it to your ~/.Renviron.",
      call. = FALSE
    )
  }
}

choicesList <- function(tableName) {
  choicesList <- list(
    person = list(
      gender_concept_id = list(
        "Male" = 8507,
        "Female" = 8532
        ),
      year_of_birth = 1900:as.integer(format(Sys.Date(), "%Y")),
      month_of_birth = 1:12,
      day_of_birth = 1:30
    )
  )
  return(choicesList[[tableName]])
}

#' Read a parquet file using DuckDB.
#' @param file Path to the parquet file.
#' @return A data frame.
#' @noRd
read_parquet_file <- function(file) {
  con <- DBI::dbConnect(duckdb::duckdb())
  on.exit(DBI::dbDisconnect(
    con,
    shutdown = TRUE
    ),
    add = TRUE
    )
  path_sql <- gsub(
    "'",
    "''",
    path.expand(file)
    )
  DBI::dbGetQuery(
    con,
    paste0(
      "SELECT * FROM read_parquet('",
      path_sql,
      "')"
      )
    )
}

columnNames <- function(
    name = NULL,
    limit = NULL,
    ommitTime = TRUE,
    cdmVersion = "5.4"
    ) {

  cdmSpecificationPath <- system.file(
    "cdmTableSpecifications",
    paste0(
      "emptycdm_",
      cdmVersion
      ),
    package = "PatientGenerator"
    )

  # supported_tables <- file.path(cdmSpecificationPath) |> 
  #    list.files() |> 
  #   stringr::str_remove(".parquet")
  
  supported_tables <- c(
    "person",
    "observation_period",
    "condition_occurrence",
    "drug_exposure",
    "measurement",
    "procedure_occurrence"
    )

  if (!is.null(name)) {

    if (length(name) != 1) {
      stop("Error: Variable 'name' must have exactly one element.")
    }

    checkmate::assertString(name)

    if (!name %in% supported_tables) {
      stop("Error: Variable 'name' should be a cdm table")
    }

    file <- file.path(
      cdmSpecificationPath,
      paste0(
        name,
        ".parquet"
        )
      )
    table_data <- read_parquet_file(file) |>
      data.table::as.data.table()
    if (isTRUE(ommitTime)) {
      table_data <- table_data[
        ,
        .SD,
        .SDcols = !grepl(
          "time",
          names(table_data)
          )
        ]
    }
    if (!is.null(limit)) {
      table_data <- table_data[, 1:limit]
    }
    return(table_data)

  } else {

    path <- file.path(cdmSpecificationPath) |> 
      list.files(
        full.names = TRUE
        )
    cdm_tables <- list()

    for(table_name in supported_tables){
      file <- file.path(
        cdmSpecificationPath,
        paste0(
          table_name,
          ".parquet"
          )
        )
      table_data <- read_parquet_file(file) |>
        data.table::as.data.table()
      if (isTRUE(ommitTime)) {
        table_data <- table_data[
          ,
          .SD,
          .SDcols = !grepl(
            "time",
            names(table_data)
            )
          ]
      }
      if (!is.null(limit)) {
        table_data <- table_data[, 1:limit]
      }
      cdm_tables[[table_name]] <- table_data
    }
    return(cdm_tables)
  }
}

retrieveCodelist <- function(concept_label = "Stage 1", domain = "Measurement") {
  codelist <- readRDS(file = system.file("concept_sets", "ovarian_cancer_codelist.rds", package = "PatientGenerator"))
  checkmate::assertCharacter(concept_label)
  checkmate::assertCharacter(domain)
  domain_names <- codelist |>
    dplyr::pull(domain_id) |>
    unique()
  if (!domain %in% domain_names) {
    warning("domain not found in codelist")
  }
  result <- codelist %>%
    filter(stringr::str_detect(concept_name, stringr::regex(concept_label, ignore_case = TRUE))) |>
    dplyr::filter(domain_id == domain) |>
    jsonlite::toJSON(dataframe = "rows",
                     auto_unbox = TRUE)
  return(result)
}

retrieveCodelistTool <- ellmer::tool(
  fun = retrieveCodelist,
  description = "Retrieves data for concept_id search by concept_name and domain_id",
  name = "retrieveCodelist",
  arguments = list(
    concept_label = ellmer::type_string(description = "A regex to look up for a concept_name; the function uses stringr::regex to filter the concept_name"),
    domain = ellmer::type_string(description = "One word typically a domain from the OMOP-CDM, currently only available now: 'Drug', 'Condition', 'Measurement', 'Observation', 'Procedure'")
  )
)

cdmConstructor <- R6::R6Class(
  "cdmConstructor",
  public = list(
    tables = NULL,
    person = NULL,
    observation_period = NULL,
    drug_exposure = NULL,
    condition_occurrence = NULL,
    measurement = NULL,
    procedure_occurrence = NULL,
    initialize = function(tables = c(
      "observation_period",
      "condition_occurrence",
      "drug_exposure",
      "measurement",
      "procedure_occurrence"
      )
    ) {
      self$tables <- tables
      self$person <- personTable$new()
      for (i in seq_along(tables)) {
        self[[tables[i]]] <- cdmTable$new(tables[i])
      }
    },
    add = function(person_id) {
      checkmate::assertInteger(person_id)
      name_event_id <- paste(
        private$.tableName,
        "id",
        sep = "_"
        )
      event_id <-  if (
        length(private$.data[[name_event_id]]) == 0) {
        1L
      } else {
          private$.data[[name_event_id]][length(
            private$.data[[name_event_id]]
            )] |> 
          as.integer() + 
          1L
      }
      new_person_data <- private$.defaultPersonData()
      new_row <- private$.constructNewRow(
        name_event_id,
        event_id, c(
          person_id = person_id,
          new_person_data
          )
        )
      private$.data <- rbindlist(
        list(
          private$.data,
          new_row
          ),
        fill = TRUE
        )
      },
    update = function(event_id, ...) {
      new_data <- list(...)
      if (!all(names(new_data) %in% names(private$.columnNames))) {
        stop(
          glue::glue(
            "Error: one column from c(
              {glue::glue_collapse(
                names(
                  new_data
                  ),
                sep = ', '
                )
              }
            ) are not from the {private$.tableName} table"
            )
          )
      }
      name_id <- paste(
        private$.tableName,
        "id",
        sep = "_"
        )
      index_table <- which(private$.data[[name_id]] == event_id)
      if (length(index_table) > 0) {
        for (col_name in names(new_data)) {
          if (stringr::str_detect(col_name, "date")) {
            new_value <- new_data[[col_name]] |> as.Date()
          } else if (stringr::str_detect(col_name, "concept") & !stringr::str_detect(col_name, "gender")) {
            new_value <- new_data[[col_name]] |> as.integer()
          } else {
            new_value <- new_data[[col_name]] |> as.character()
          }
          data.table::set(
            private$.data,
            i = index_table,
            j = col_name,
            value = new_value
            )
        }
      } else {
        warning(glue::glue("{name_id} not found in table"))
      }
      },
    extractRow = function(event_id) {
      name_event_id <- paste(
        private$.tableName,
        "id",
        sep = "_"
        )
      private$.data[private$.data[[name_event_id]] == event_id, ]
    },
    load = function(jsonData) {
      private$.data <- jsonData
    },
    delete = function(event_id) {
      name_id <- paste(
        private$.tableName,
        "id",
        sep = "_"
        )
      index_table <- which(private$.data[[name_id]] == event_id)
      private$.data <- private$.data[-index_table]
    },
    # Add data from json test set
    loadJsonTestSet = function(path) {
      self$reset()
      # cdm_schema <- jsonvalidate::json_validator(
      #   system.file("jsonSchemas",
      #               "cdm54schema-complete.json",
      #               package = "TestGenerator"),
      #   engine = "ajv"
      #   )
      checkmate::assertFileExists(path)
      jsonData <- jsonlite::fromJSON(path)
      # if (!cdm_schema(toJSON(jsonData, auto_unbox = TRUE))) {
      #   stop("Invalid data structure!")
      # }
      currentTables <- names(jsonData)
      # Check for the expected columns in the CDM
      for (tableName in currentTables) {
        if (tableName %in% c("person",
                             "observation_period",
                             "condition_occurrence",
                             "drug_exposure",
                             "measurement",
                             "procedure_occurrence")) {
          classTable <- class(jsonData[[tableName]])
          table_data <- jsonData[[tableName]] |> as.data.table()
          if (classTable == "data.frame") {
              date_cols <- names(table_data)[grepl("_date$", names(table_data))]
              if (length(date_cols) > 0 ) {
                table_data[, (date_cols) := lapply(.SD, as.Date), .SDcols = date_cols]
              }
            self[[tableName]]$load(data.table::rbindlist(list(private$.data, table_data)))
          }
        }
      }
    },
    # Reset
    reset = function() {
      self$person$reset()
      self$observation_period$reset()
      self$drug_exposure$reset()
      self$condition_occurrence$reset()
      self$measurement$reset()
      self$procedure_occurrence$reset()
    },

    # Delete person or event
    deletePersonEvent = function(event_id, type) {
      table_data <- glue::glue("{type}_data")
      table_column <- glue::glue("{type}_id")
      data_table <- self[[table_data]]
      self[[table_data]] <- data_table[data_table[[table_column]] != event_id, ]
    },

    # Update Dates
    updateDates = function(person_id,
                           event_id,
                           start_date,
                           end_date) {
      start_date <- as.Date(start_date)
      end_date <- as.Date(end_date)
      name_id <- private$.tableNameId()
      name_start_date <- private$.tableNameDate("start")
      name_end_date <- private$.tableNameDate("end")
      index_table <- which(private$.data[[name_id]] == event_id)
      if (length(index_table) > 0) {
        if (!is.null(start_date)) {
          data.table::set(
            private$.data,
            i = index_table,
            j = name_start_date,
            value = start_date
            )
        }
        if (!is.null(end_date)) {
          data.table::set(
            private$.data,
            i = index_table,
            j = name_end_date,
            value = end_date)
        }
      } else {
        warning(
          glue::glue(
            "{name_id} not found in table"
            )
          )
      }
      },
    
    # Export data to json
    getCdmData = function() {
      cdm_data <- list(
        person = self$person$data(),
        observation_period = self$observation_period$data(),
        drug_exposure = self$drug_exposure$data(),
        condition_occurrence = self$condition_occurrence$data(),
        procedure_occurrence = self$procedure_occurrence$data()
        )

        cdm_data_json <- jsonlite::toJSON(
          cdm_data,
          dataframe = "rows",
          pretty = TRUE,
          na = "null")
        
        return(cdm_data_json)
        
        },
    # Export data to json
    getCdmDataTimeline = function() {
      if (self$person$data() |> length() > 0) {
        tables_timeline <- setNames(
          lapply(
            self$tables, 
            function(table_name) {
              cols <- c(
                self[[table_name]]$tableNameId(),
                "person_id",
                self[[table_name]]$tableNameConceptId(),
                self[[table_name]]$tableNameDate("start"),
                self[[table_name]]$tableNameDate("end")
              )
              table_data <- self[[table_name]]$data() 
              dt_timeline <- data.table::copy(
                table_data[, ..cols]
                ) 
              if (length(cols) == 5) {
                data.table::setnames(
                  dt_timeline,
                  old = cols,
                  new = c(
                    "event_id",
                    "person_id",
                    "concept_id",
                    "start_date",
                    "end_date"
                  )
                )
              } else {
                data.table::setnames(
                  dt_timeline,
                  old = cols,
                  new = c(
                    "event_id",
                    "person_id",
                    "concept_id",
                    "start_date"
                  )
                )
              }
            }
          ),
          basename(self$tables)
        )
        
        data_timeline <- rbindlist(
          tables_timeline,
          fill = TRUE,
          idcol = "type"
        ) |> 
          mutate(categories = row_number())
        
        } else {
          data_timeline <- data.table::data.table()
          }
      return(data_timeline)
      },
    # Export data to json
    getDrugExposureData = function() {
      drug_exposure_json <- jsonlite::toJSON(
        self$drug_exposure_data,
        dataframe = "rows",
        null = "null",
        na = "null",
        auto_unbox = TRUE,
        digits = getOption(
          "shiny.json.digits",
          16
          ),
        use_signif = TRUE,
        force = TRUE,
        POSIXt = "ISO8601",
        UTC = TRUE,
        rownames = FALSE,
        keep_vec_names = TRUE,
        json_verabitm = TRUE
        )
      return(drug_exposure_json)
      }
    ),
  private = list(
    .getData = function() {
      return(private$.data)
      },
    .emptyTable = function() {
      private$.columnNames
      },
    .defaultPersonData = function() {
      column_names <- private$.columnNames |>
        names() |>
        tail(-2) |>
        head(3)
      if (private$.tableName == "observation_period") {
        values <- list(
          as.Date("2010-02-28"),
          as.Date("2015-02-28"),
          44191562L
          )
        } else {
          values <- list(
            44191562L,
            as.Date("2010-02-28"),
            as.Date("2015-02-28")
            )
          }
      person_data <- Map(function(column_name, value) {
        value
        },
        column_names,
        values
        )
      return(person_data)
      },
    .constructNewRow = function(name_event_id, event_id, ...) {
      checkmate::assertCharacter(name_event_id)
      checkmate::assertInteger(event_id)
      new_data <- c(
        setNames(
          list(event_id),
          name_event_id),
        ...
        )
      empty_row <- private$.emptyTable()
      new_row <- rbindlist(
        list(
          empty_row,
          new_data
          ),
        fill = TRUE
        )
      return(new_row)
      }
    )
)

test_that("cdmTable condition_occurrence supports add/reset/delete", {
  condition_tbl <- new_cdm_table("condition_occurrence")
  condition_tbl$add(person_id = 1L)
  condition_tbl$add(person_id = 2L)

  dat <- condition_tbl$data()
  expect_equal(dat$condition_occurrence_id, c(1L, 2L))
  expect_equal(dat$person_id, c(1L, 2L))

  condition_tbl$delete(event_id = 2L)
  dat <- condition_tbl$data()
  expect_equal(dat$condition_occurrence_id, 1L)

  condition_tbl$reset()
  expect_length(condition_tbl$data()$condition_occurrence_id, 0)
})

test_that("Initialize correctly tables defined in the parameter", {
  
  tables <- c(
    "observation_period"
  )
  
  cdm <- cdmConstructor$new(tables = tables)
  
  for (i in seq_along(tables)) {
    expect_r6_class(
      cdm[[tables[i]]],
      "cdmTable"
    )
  }
  
  tables <- c(
    "observation_period",
    "drug_exposure",
    "condition_occurrence",
    "measurement",
    "procedure_occurrence"
    )
  
  cdm <- cdmConstructor$new(tables = tables)
  
  for (i in seq_along(tables)) {
    expect_r6_class(
      cdm[[tables[i]]],
      "cdmTable"
    )
  }
  
})

test_that("cdmConstructor reset empties core tables", {
  
  cdm <- new_cdm()
  
  cdm$person$add(
    gender_concept_id = 8532L,
    year_of_birth = 1967L
    )
  
  # Add 1 person to all data tables
  cdm_tables <- c(
    "observation_period",
    "condition_occurrence",
    "drug_exposure",
    "measurement",
    "procedure_occurrence"
  )
  
  for (i in seq_along(cdm_tables)) {
    expect_no_error({
      cdm[[cdm_tables[i]]]$add(
        person_id = 1L
        )
    })
  }
  
  # Check info was added to all tables
  for (i in seq_along(cdm_tables)) {
    cdm[[cdm_tables[i]]]$data() |> 
      pull(person_id) |> 
      length() |> 
      expect_equal(1)
  }
  
  ###
  cdm$reset()
  
  # Empty data after reset
  # Check info was added to all tables
  for (i in seq_along(cdm_tables)) {
    cdm[[cdm_tables[i]]]$data() |> 
      pull(person_id) |> 
      length() |> 
      expect_equal(0)
  }

})

test_that("person table add/update/delete behaves as expected", {
  
  cdm <- new_cdm()
  
  # ADD
  cdm$person$add(
    gender_concept_id = 8532L,
    year_of_birth = 1967L
    )
  cdm$person$add(
    gender_concept_id = 8507L,
    year_of_birth = 1988L
    )
  
  # UPDATE
  cdm$person$update(
    person_id = 1L,
    gender_concept_id = 8507L,
    year_of_birth = 1977L
    )

  # TEST
  person_table <- cdm$person$data()
  expect_equal(person_table$person_id, c(1L, 2L))
  expect_equal(person_table$gender_concept_id, c(8507L, 8507L))
  expect_equal(person_table$year_of_birth, c(1977L, 1988L))

  # DELETE
  cdm$person$delete(person_id = 2L)
  person_table <- cdm$person$data()
  expect_equal(person_table$person_id, 1L)
})

test_that("observation/condition/drug/measurement/procedure_occurrence update dates", {
  cdm <- new_cdm()
  cdm$person$add(gender_concept_id = 8532L, year_of_birth = 1967L)
  
  cdm_tables <- c(
    "observation_period",
    "condition_occurrence",
    "drug_exposure",
    "measurement",
    "procedure_occurrence"
  )
  
  # Add 1 person to all data tables
  for (i in seq_along(cdm_tables)) {
    expect_no_error({
      cdm[[cdm_tables[i]]]$add(
        person_id = 1L
      )
    })
  }
  
  # Default dates
  for (i in seq_along(cdm_tables)) {

    # Check default start date
    cdm[[cdm_tables[i]]]$data()[,.SD, .SDcols = patterns("date")] |> 
      select(1) |> 
      pull() |>
      as.character() |> 
      expect_equal("2010-02-28")
    
    # Update dates
    expect_no_error({  
      cdm[[cdm_tables[i]]]$updateDates(
        person_id = 1L,
        event_id = 1L,
        start_date = as.Date("1997-10-29"),
        end_date = as.Date("1999-10-29")
      )
    })
      
      # Check new date
    cdm[[cdm_tables[i]]]$data()[,.SD, .SDcols = patterns("date")] |> 
      select(1) |> 
      pull() |>
      as.character() |> 
      expect_equal("1997-10-29")
    
  }
      
})

test_that("getCdmData and getCdmDataTimeline return valid structures", {
  cdm <- new_cdm()
  cdm$person$add(gender_concept_id = 8532L, year_of_birth = 1967L)
  
  cdm_tables <- c(
    "observation_period",
    "condition_occurrence",
    "drug_exposure",
    "measurement",
    "procedure_occurrence"
  )
  
  # Add 1 person to all data tables
  for (i in seq_along(cdm_tables)) {
    expect_no_error({
      cdm[[cdm_tables[i]]]$add(
        person_id = 1L
      )
    })
  }

  cdm_json <- cdm$getCdmData()
  expect_true(
    jsonlite::validate(cdm_json)
    )

  timeline <- cdm$getCdmDataTimeline()
  expect_s3_class(
    timeline,
    "data.table"
    )
  expect_in(
    names(timeline),
    c("event_id",
      "concept_id",
      "person_id",
      "start_date",
      "end_date",
      "type",
      "categories")
    )
  expect_in(
    unique(timeline$type),
    c("observation_period",
      "drug_exposure",
      "condition_occurrence",
      "measurement",
      "procedure_occurrence"
      )
  )
})

test_that("loadJsonTestSet loads source test fixtures", {
  cdm <- new_cdm()
  path <- testthat::test_path("testCases", "objective_1_patients.json")
  expect_no_error(cdm$loadJsonTestSet(path))
  expect_gt(nrow(cdm$person$data()), 0)
})

test_that("Testing methods on LLM testset", {
  testthat::skip_on_cran()
  
  # # An LLM testset for this test
  # model <- pick_openai_model()
  # patientGenerator <- patientChat$new(model = "gpt-5.4-mini")
  # 
  # ### Test set description for this test:
  # patientGenerator$prompt(
  # "Population (person table):
  #   - 10 adult patients
  #   - 5 female
  #   - 5 male
  # 
  #  Observation Period:
  #   - Start date between date of birth each person and end of observation 2025-12-31
  # 
  #  Condition Occurrence:
  #    - All patients must have Diabetes (condition_concept_id: 201826)
  #    - Condition start date between 2015-01-01 and 2020-12-31
  # 
  #  Drug Exposure:
  #    - All patients must have Semaglutide (drug_concept_id: 19079450)
  #    - Drug exposure in a window of 0 to 30 days after index date
  # 
  #  Measurement:
  #    - All patients must have Fasting glucose (measurement_concept_id: 3018251)
  # 
  #  Procedure cccurrence:
  #    - 50% of patients (5 patients) must have Amputation of toe (procedure_concept_id: 4159766)
  # 
  #  Output Requirements:
  #   - Fill only specified tables in this prompt"
  #   )
  # patientGenerator$save("test_diabetes_patients")

  ### Check testset
  cdm <- TestGenerator::patientsCDM(
    testName = "test_diabetes_patients",
    cdmVersion = "5.4"
    )

  cdm$person |>
    collect() |>
    nrow() |>
    expect_equal(10)
  
  cdm$procedure_occurrence |>
    collect() |>
    nrow() |>
    expect_equal(5)
  
  path <- testthat::test_path(
    "testCases",
    "test_diabetes_patients.json"
    )
  
  cdm_tables <- c(
    "observation_period",
    "condition_occurrence",
    "drug_exposure",
    "measurement",
    "procedure_occurrence"
  )
  
  expect_no_error({
    cdm <- new_cdm()
    cdm$loadJsonTestSet(path)
    for (i in seq_along(cdm_tables)) {
      expect_no_error({
        cdm[[cdm_tables[i]]]$add(
          person_id = 1L
        )
      })
    }
  })
  
  cdm$condition_occurrence$data() |> 
    pull(person_id) |> 
    expect_length(11)
  
  cdm$drug_exposure$data() |> 
    pull(person_id) |> 
    expect_length(31)
  
  cdm$measurement$data() |> 
    pull(person_id) |> 
    expect_length(11)
  
  cdm$procedure_occurrence$data() |> 
    pull(person_id) |> 
    expect_length(6)
  
})

test_that("Testing modified test from LLM can be inserted back to TestGenerator", {
  testthat::skip_on_cran()
  
  # Using the same diabetes test from previous test
  path <- testthat::test_path(
    "testCases",
    "test_diabetes_patients.json"
  )

  cdm_tables <- c(
    "observation_period",
    "condition_occurrence",
    "drug_exposure",
    "measurement",
    "procedure_occurrence"
  )  
  # Load into memory and modify all fields
  expect_no_error({
    cdm <- new_cdm()
    cdm$loadJsonTestSet(path)
  
    for (i in seq_along(cdm_tables)) {
      expect_no_error({
        cdm[[cdm_tables[i]]]$add(
          person_id = 1L
        )
      })
    }
    
    # Update dates
    cdm[[cdm_tables[i]]]$updateDates(
      person_id = 1L,
      event_id = 1L,
      start_date = as.Date("1997-10-29"),
      end_date = as.Date("1999-10-29")
    )
  })
  
  mod_test_file <- testthat::test_path(
    "testCases",
    "mod_test_file.json"
  )
  
  write(
    cdm$getCdmData(),
    file = mod_test_file
  )
  
  expect_no_error({
    cdm <- TestGenerator::patientsCDM(
      testName = "mod_test_file",
      cdmVersion = "5.4"
      )

    cdm$person |>
      collect() |>
      nrow() |>
      expect_equal(10)

    cdm$procedure_occurrence |>
      collect() |>
      nrow() |>
      expect_equal(6)
  })
  
  unlink(mod_test_file, recursive = TRUE)

})

test_that("Testing methods on LLM testset 'objective_1_patient'", {
  
  path <- testthat::test_path(
    "testCases",
    "objective_1_patients.json"
  )
  
  cdm_tables <- c(
    "observation_period",
    "condition_occurrence",
    "drug_exposure",
    "measurement",
    "procedure_occurrence"
  )
  
  expect_no_error({
    cdm <- new_cdm()
    cdm$loadJsonTestSet(path)
    for (i in seq_along(cdm_tables)) {
      expect_no_error({
        cdm[[cdm_tables[i]]]$add(
          person_id = 1L
        )
      })
    }
  })
  
  cdm$condition_occurrence$data() |> 
    pull(person_id) |> 
    expect_length(5)
  
  cdm$drug_exposure$data() |> 
    pull(person_id) |> 
    expect_length(5)
  
  cdm$measurement$data() |> 
    pull(person_id) |> 
    expect_length(11)
  
  cdm$procedure_occurrence$data() |> 
    pull(person_id) |> 
    expect_length(1)
  
})

test_that("cdmTableServer add action appends event for selected person", {
  cdm <- new_cdm()
  cdm$person$add(gender_concept_id = 8532L, year_of_birth = 1970L)
  cdm$observation_period$add(person_id = 1L)

  syncing <- shiny::reactiveVal(FALSE)

  shiny::testServer(cdm_table_server,
    args = list(
      id = "observation_period",
      cdm = cdm,
      person_id_selected = shiny::reactive("1"),
      syncing = syncing
    ), {
      session$setInputs(add = 1)
    }
  )

  expect_equal(nrow(cdm$observation_period$data()), 2L)
  expect_equal(tail(cdm$observation_period$data()$person_id, 1), 1L)
})

personUI <- function(id) {
  ns <- NS(id)
  card(
    # full_screen = TRUE,
    card_body(
      fluidRow(column(2, actionButton(ns("create"), "Create")),
               column(2, actionButton(ns("duplicate"), "Duplicate")),
               column(2, actionButton(ns("delete"), "Delete"))),
       br(),
       fluidRow(
         tagList(
           createInputs(ns,
                        type = "person",
                        columns = columnNames(name = id, limit = 5) |> names())
          )
        ),
      # div(
      #   class = "reveal-on-full-screen",
      #   h5("Other columns"),
      #   fluidRow(
      #     tagList(createInputs(ns,
      #                          type = "person",
      #                          columns = c("person_id", "gender_concept_id", "year_of_birth"),
      #                          inverse = TRUE))
      #     )
      #   ),
      fillable = FALSE
      )
  )
  }

personServer <- function(id, cdm, trigger = NULL) {

  moduleServer(id, function(input, output, session) {
    ns <- session$ns

    if (!is.null(trigger)) {
      observeEvent(trigger(), {
        req(trigger() > 0)

        personTable <- cdm$person$data()

        new_choices <- personTable$person_id
        new_selected <- if (length(new_choices) > 0) new_choices[1] else NULL

        updateSelectInput(session, "person_id",
                          choices = new_choices,
                          selected = new_selected)

      })
    }
    # ------------------------------------------

    # CREATE
    observeEvent(input$create, {
      # Randomly select persons features
      gender_concept_id <- sample(c(8532, 8507), 1) # 8532 is Female, 8507 is male
      year_of_birth <- sample(1920:2004, 1)
      month_of_birth <- sample(seq(1, 12), 1)
      day_of_birth <- sample(seq(1, 31), 1)
      cdm$person$add(gender_concept_id = gender_concept_id,
                     year_of_birth = year_of_birth,
                     month_of_birth = month_of_birth,
                     day_of_birth = day_of_birth)
      personTable <- cdm$person$data()
      updateSelectInput(session, "person_id", choices = personTable$person_id, selected = personTable$person_id[length(personTable$person_id)])
    })

    # DUPLICATE
    observeEvent(input$duplicate, {
      cdm$person$add(gender_concept_id = input$gender_concept_id,
                        year_of_birth = input$year_of_birth,
                        month_of_birth = input$month_of_birth)
      personTable <- cdm$person$data()
      updateSelectInput(session, "person_id", choices = personTable$person_id, selected = personTable$person_id[length(personTable$person_id)])
    })

    # DELETE
    observeEvent(input$delete, {
      req(input$person_id)
      cdm$person$delete(person_id = as.integer(input$person_id))
      personTable <- cdm$person$data()
      # Update to last person before the deletion
      if (personTable$person_id %>% length() == 0) {
        # Update to null - If not, the value of the previous person remains after the person id is deleted
        updateSelectInput(session, "person_id", choices = personTable$person_id, selected = personTable$person_id[1])
        # ids <- c("gender_concept_id", "year_of_birth", "month_of_birth", "day_of_birth")
        # Update gender and year of birth for only that person
        updatePersonInputs(cdmPersonData = cdm$person$data(),
                           input_person_id = input$person_id,
                           ids = columnNames(name = id, limit = 5) |> names() |> tail(-1),
                           session = session)
      } else {
        updateSelectInput(session, "person_id", choices = personTable$person_id, selected = personTable$person_id[as.numeric(input$person_id) - 1])
      }

    })

    # UPDATE INTERFACE
    observeEvent(input$person_id, {
      req(input$person_id)
      ids <- setdiff(columnNames(name = "person") |> names(), "person_id")
      updatePersonInputs(cdmPersonData = cdm$person$data(),
                         input_person_id = input$person_id,
                         ids = ids[1:4],
                         session = session)

    }, ignoreInit = TRUE)

    # UPDATE DATA
    observeEvent({
      input$gender_concept_id
      input$year_of_birth
      input$month_of_birth
      input$day_of_birth
    }, {
      req(input$gender_concept_id,
          input$year_of_birth,
          input$month_of_birth,
          input$day_of_birth)
      cdm$person$update(person_id = input$person_id |> as.integer(),
                        gender_concept_id = input$gender_concept_id |> as.integer(),
                        year_of_birth = input$year_of_birth |> as.integer(),
                        month_of_birth = input$month_of_birth |> as.integer(),
                        day_of_birth = input$day_of_birth |> as.integer())

    }, ignoreInit = TRUE)

    return(reactive(input$person_id))

  })
}

#' `patientChat()` generates synthetic patients in the OMOP-CDM using an LLM API.
#'
#' @description
#' Requires an OPEN_AI_KEY in ~/.Renviron.
#' After that just sent a prompt and save() the results.
#' The JSON file can be used as an OMOP-CDM patient test set.
#' @details
#' Accepts a prompt as input. Produces a test set using a
#' structured JSON schema. Utilizes tools such as CodelistGenerator or Hecate
#' to look up concept IDs. Accepts subsequent prompts to modify existing test
#' sets that the LLM uses as context.
#'
#' This class allows testing patient sets created by the LLM, prompt engineering,
#' integration of search tools and functionality, and creating a set of patients
#' to test analytical packages.
#' @returns 
#' A JSON response that includes: the natural language answer from the LLM and
#' a JSON with test set patients in accordance to the provided schema.
#' @importFrom ellmer chat_openai
#' @importFrom ellmer type_from_schema
#' @importFrom jsonlite fromJSON
#' @importFrom httr2 request req_headers req_perform resp_body_json
#' @importFrom checkmate assertCharacter assertFileExists assertDirectoryExists
#' @importFrom tools file_path_sans_ext
#' @importFrom R6 R6Class
#' @export
#' @examples
#' \dontrun{
#' generator <- patientChat$new()
#' generator$prompt("Give me 5 patients")
#' generator$save("my_test")
#' }
patientChat <- R6::R6Class(
  classname = "patientChat",
  public = list(

    #' @field chat An ellmer chat instance
    chat = NULL,

    #' @field json_schema_path JSON schema to output structured results
    json_schema_path = NULL,

    #' @field response Ouput from the LLM
    response = NULL,

    #' @field codelist A codelist with details to search for concepts ids
    codelist = NULL,

    #' @description
    #' Create a new chat to create JSON test sets for OMOP-CDM.
    #' @param system_prompt Initial system prompt to impose behaviour to the LLM
    #' @param model Such as "gpt-5.3". For a complete list, call patientChat$availableModels()
    #' @param jsonSchemaPath The JSON schema to structure output from LLM
    #' @param echo How the output will be displayed in the console
    #' @param codelist_data A codelist with details to search for concepts ids
    #'
    #' @return A new `Person` object.
    initialize = function(system_prompt = NULL,
                          model = "gpt-5.4",
                          jsonSchemaPath = NULL,
                          echo = c("none", "output", "all"),
                          codelist_data = NULL) {

      # Check API and available models -----------------------
      private$.api_check(model)

      # check JSON schema file -------------------------------
      private$.json_schema_check(jsonSchemaPath)

      # System propmpt ---------------------------------------
      if (is.null(system_prompt)) {
        system_prompt <- "You are an automatic generator of synthetic patients data using exclysively the OMOP-CDM v5.4 format"
      }

      # Create chat ------------------------------------------
      self$chat <- ellmer::chat_openai(system_prompt = system_prompt,
                                       model = model,
                                       echo = echo)
      message("Chat created")

      # Codelist ---------------------------------------------
      if (!is.null(codelist_data)) {
        checkmate::assertDataFrame(codelist_data)
        if (all(c(
          "concept_id",
          "concept_name",
          "domain_id",
          "vocabulary_id",
          "standard_concept"
          ) %in% codelist_data |> 
          names()
          )
          ) {
          self$codelist <- codelist_data
          self$chat$register_tool(
            private$.register_codelist_tool()
            )
          message("Codelist tool added")
        } else {
          stop(
            "Dataframe has missing columns: 'concept_id', 'concept_name', 'domain_id', 'vocabulary_id', 'standard_concept'"
            )
        }
      }

    },

    #' @description
    #' Prompt to request data from LLM API
    #' @param prompt A query in character.
    prompt = function(prompt) {
      checkmate::assertCharacter(prompt)
      api_respose <- self$chat$chat_structured(
        prompt,
        type = ellmer::type_from_schema(
          path = self$json_schema_path
        )
      )
      self$response <- api_respose
    },

    #' @description
    #' Output in JSON format
    json_response = function() {
      jsonlite::toJSON(
        self$response,
        dataframe = "rows",
        pretty = TRUE,
        null = "null",
        na = "null",
        auto_unbox = TRUE
        )
    },

    #' @description
    #' Returns the chat object
    output = function() {
      return(self$chat)
    },

    #' @description
    #' Retrieves and filters data from codelist_data
    #' @param concept_label Filters the concept_name in the codelist with details
    #' @param domain Filters the domain in the codelist with details.
    retrieveCodelist = function(concept_label = "Stage 1", domain = "Measurement") {
      checkmate::assertCharacter(concept_label)
      checkmate::assertCharacter(domain)
      domain_names <- self$codelist |>
        dplyr::pull(domain_id) |>
        unique()
      if (!domain %in% domain_names) {
        warning("domain not found in codelist")
      }
      result <- self$codelist %>%
        filter(stringr::str_detect(
          concept_name,
          stringr::regex(
            concept_label,
            ignore_case = TRUE
            )
          )
          ) |>
        dplyr::filter(domain_id == domain) |>
        jsonlite::toJSON(
          dataframe = "rows",
          auto_unbox = TRUE
          )
      return(result)
    },

    #' @description
    #' Saves the JSON test set to disk.
    #' @param name Name of the file
    #' @param path To save the file.
    #' If NULL, the package first tries `testthat::test_path("testCases")`,
    #' then checks `options(PatientGenerator.testSetDir = "...")`, and finally
    #' falls back to the package user data directory.
    save = function(name = "patient-chat-test",
                    path = NULL) {
      if (is.null(path)) {
        path <- testSetDir(create = TRUE)
      } else {
        checkmate::assertCharacter(path)
        checkmate::assertDirectoryExists(path)
      }
      name <- tools::file_path_sans_ext(name)
      test_file_path <- file.path(
        path,
        paste0(
          name,
          ".json"
          )
        )
      write(
        self$json_response(),
        file = test_file_path
        )
    },

    #' @description
    #' Retrieves available models from the LLM API.
    availableModels = function() {
      private$.check_api_key()
      # Retrieve and check available models
      response <- httr2::request("https://api.openai.com/v1/models") |>
        httr2::req_headers(
          Authorization = paste(
            "Bearer",
            Sys.getenv("OPENAI_API_KEY")
            )
        ) |>
        httr2::req_perform()
      models <- httr2::resp_body_json(response)
      lapply(models$data, function(data) data$id) |>
        unlist()
    }
  ),

  private = list(

    .api_check = function(model) {
      checkmate::assertCharacter(model)
      api_models <- self$availableModels()
      if (!model %in% api_models) {
        stop(
          glue::glue("{model} not available.\n"),
          "\n These are some models that are available to you:\n",
          paste0("  - ", sample(api_models, 10), collapse = "\n"),
          "\n For a complete list, call availableModels()\n",
          call. = FALSE
        )
      } else {
        message(glue::glue("{model} is a valid model"))
      }
    },

    .json_schema_check = function(jsonSchemaPath) {
      if (is.null(jsonSchemaPath)) {
        json_schema_path <- system.file(
          "jsonSchemas",
          "cdm54schema-complete.json",
          package = "PatientGenerator"
          )
        checkmate::assertFileExists(json_schema_path)
        self$json_schema_path <- json_schema_path
        } else {
        checkmate::assertCharacter(jsonSchemaPath)
        checkmate::assertFileExists(jsonSchemaPath)
        is_json <- tryCatch({
          jsonlite::fromJSON(jsonSchemaPath)
          message("The file at 'jsonSchemaPath' can be parsed as JSON")
          self$json_schema_path <- jsonSchemaPath
        }, error = function(e) {
          stop("The file at 'jsonSchemaPath' cannot be parsed as JSON")
        })
      }
    },

    .check_api_key = function(apiKeyName = "OPENAI_API_KEY") {
      key <- Sys.getenv("OPENAI_API_KEY", unset = "")
      if (!nzchar(key)) {
        stop(
          "API key not found.\n",
          "Set it with Sys.setenv(OPENAI_API_KEY = 'your_key') ",
          "and/or add it to your ~/.Renviron.",
          call. = FALSE
        )
      }
    },

    .register_codelist_tool = function() {
      retrieveCodelistTool <- ellmer::tool(
        fun = self$retrieveCodelist,
        description = "Retrieves data for concept_id search by concept_name and domain_id",
        name = "retrieveCodelist",
        arguments = list(
          concept_label = ellmer::type_string(
            description = "A regex to look up for a concept_name; the function uses stringr::regex to filter the concept_name"
            ),
          domain = ellmer::type_string(
            description = "One word typically a domain from the OMOP-CDM, currently only available now: 'Drug', 'Condition', 'Measurement', 'Observation', 'Procedure'"
            )
        )
      )
    }
  )
)

test_that("Diabetes simple", {
  skip_if_no_openai()
  
  expect_no_error({
    cdm <- TestGenerator::patientsCDM(
      testName = "diabetes_simple",
      cdmVersion = "5.4"
    )
    cdm$person
    cdm$condition_occurrence
    cdm$drug_exposure
  })
  
})

test_that("Diabetes workflow prompt", {
  skip_if_no_openai()
  
  # Get a glimpse of all available models 
  models <- PatientGenerator::availableModels()
  
  # Create a chat instance 
  patientGenerator <- patientChat$new(model = "gpt-5.5")

  # Prompt a detailed description of test patients
  patientGenerator$prompt(
  "Population (person table):
    - 10 adult patients
    - 5 female, use gender_concept_id = 8532
    - 5 male, use gender_concept_id = 8507

   Observation Period:
    - Start date between date of birth each person and end of observation 2025-12-31

   Condition Occurrence:
     - All patients must have Diabetes (condition_concept_id: 201826)
     - Condition start date between 2015-01-01 and 2020-12-31

   Drug Exposure:
     - All patients must have three Semaglutide (drug_concept_id: 19079450).
     - Drug exposure in a window of 0 to 30 days after index date

   Measurement:
     - All patients must have Fasting glucose (measurement_concept_id: 3018251)

   Procedure cccurrence:
     - 50% of patients (5 patients) must have Amputation of toe (procedure_concept_id: 4159766).

   Output Requirements:
    - Fill only specified tables in this prompt
    - All patients in person have an observation period
    - Fill out end dates in every table where you can"
    )
  
  # Save patients 
  patientGenerator$save("test_diabetes_patients")
  
  # JSON ready to load into TestGenerator and create a CDM reference
  cdm <- TestGenerator::patientsCDM(
    testName = "test_diabetes_patients",
    cdmVersion = "5.4"
  )
  
  cdm$person |> 
    collect() |> 
    nrow() |> 
    expect_equal(10)
  
  # Test number of females
  cdm$person |> 
    collect() |> 
    dplyr::filter(gender_concept_id == 8532) |> 
    nrow() |> 
    expect_equal(5)
  
  cdm$condition_occurrence |>
    collect() |>
    pull(condition_start_date) |>
    (\(x) all(x > as.Date("2015-01-01"), na.rm = TRUE))() |> 
    expect_true()
  
 cdm$condition_occurrence |>
    collect() |>
    pull(condition_start_date) |>
    (\(x) all(x < as.Date("2020-12-31"), na.rm = TRUE))() |> 
    expect_true()
 
 expect_no_error({
   cdm$diabetes_semaglutide <- CohortConstructor::conceptCohort(
     cdm = cdm,
     conceptSet = list(
       "diabetes" = 201826L,
       "semaglutide" = 19079450L
       ),
     name = "diabetes_semaglutide",
     exit = "event_end_date"
   )
 })
 
 expect_no_error({
   cdm$diabetes_semaglutide |> 
     CohortCharacteristics::summariseCharacteristics() |> 
     CohortCharacteristics::tableCharacteristics(
       type = "flextable",
       style = "darwin"
     )
 })
 
 expect_no_error({
   cdm$diabetes_semaglutide |> 
     CohortCharacteristics::summariseCharacteristics(
       cohortId = 1, #diabetes
       cohortIntersectFlag = list(
         targetCohortTable = "diabetes_semaglutide",
         targetCohortId = "semaglutide",
         window = c(0, 30)
       )
     ) |> 
     CohortCharacteristics::tableCharacteristics(
       type = "flextable",
       style = "darwin"
     )
 })
 
 patientGenerator$prompt(
   "Within the current records in the drug exposure tables, each drug exposure should be 30 days long"
 )
 
 patientGenerator$save("test_diabetes_patients_30")
 
 cdm <- TestGenerator::patientsCDM(
   testName = "test_diabetes_patients_30",
   cdmVersion = "5.4"
 )
 
 cdm$drug_exposure %>% 
   select(person_id,
          drug_exposure_start_date,
          drug_exposure_end_date) %>%
   mutate(days = !!CDMConnector::datediff(
     "drug_exposure_start_date",
     "drug_exposure_end_date"
     )
     ) %>%
   pull(days) %>% 
   unique() %>% 
   expect_equal(29)
 
})

test_that("Ovarian cancer stages", {
  skip_if_no_openai()
  
  patientGenerator <- PatientGenerator::patientChat$new()
  
  patientGenerator$prompt({
    "Five females (all over 18 years old) have an observation period from 2000 to 2024.

     All five have a condition occurrence of ovarian cancer (concept ID: 200051) recorded on 2012-01-01.

     Cancer stage information is recorded in the **measurement** table as follows:

      - **Female 1**:
      - Stage 1 (concept ID: 1633306) on 2012-01-01
      - Stage 2 (concept ID: 1634209) on 2012-01-02

      - **Female 2**:
      - Stage 2 (concept ID: 1634209) on 2012-01-01
      - Stage 3 (concept ID: 1633650) on 2012-01-02

      - **Female 3**:
      - Stage 3 (concept ID: 1633650) on 2012-01-01
      - Stage 4 (concept ID: 1634766) on 2012-01-02

      - **Female 4**:
      - Stage 1 (concept ID: 1633306) recorded (date not specified)

      - **Female 5**:
      - No cancer stage measurement record"
  })
  
  patientGenerator$save("patient_chat_ovarian_stages")
  
  expect_no_error({
    cdm <- TestGenerator::patientsCDM(
      testName = "patient_chat_ovarian_stages",
      cdmVersion = "5.4"
    )
  })
  
})

test_that("patientChat benchmark across all available models for diabetes + semaglutide cohort", {
  # ---------------------------------------------------------------------------
  # PLAN (manual benchmark, disabled by default):
  # 1) List all available models via availableModels().
  # 2) For each model, run patientChat generation with a fixed representative task:
  #    - exactly 10 synthetic patients
  #    - diabetes condition_occurrence
  #    - semaglutide drug_exposure
  # 3) Measure elapsed time per model.
  # 4) Validate generated JSON by loading it with cdmConstructor.
  # 5) Save benchmark results to CSV for later comparison.
  # ---------------------------------------------------------------------------

  skip_if_no_openai()
  testthat::skip_on_cran()

  run_benchmark <- tolower(Sys.getenv("PATIENTGENERATOR_RUN_MODEL_BENCHMARK", "false")) %in%
    c("1", "true", "yes")
  if (!run_benchmark) {
    testthat::skip(
      "Benchmark disabled. Set PATIENTGENERATOR_RUN_MODEL_BENCHMARK=true to run."
    )
  }

  models <- unique(PatientGenerator::availableModels())
  if (length(models) == 0) {
    testthat::skip("No models returned by availableModels().")
  }

  # Optional filtering for targeted benchmark runs.
  # Example:
  #   PATIENTGENERATOR_MODEL_BENCHMARK_REGEX='^(gpt-5|gpt-5-mini|gpt-5-nano)$'
  model_regex <- Sys.getenv("PATIENTGENERATOR_MODEL_BENCHMARK_REGEX", unset = "")
  if (nzchar(model_regex)) {
    models <- models[grepl(model_regex, models)]
  }

  if (length(models) == 0) {
    testthat::skip("No models matched PATIENTGENERATOR_MODEL_BENCHMARK_REGEX.")
  }

  benchmark_prompt <- paste(
    "Generate exactly 10 synthetic patients in OMOP-CDM v5.4.",
    "For each patient, include at least one condition_occurrence for diabetes mellitus.",
    "For each patient, include at least one drug_exposure for semaglutide.",
    "Return valid JSON following the provided schema."
  )

  sanitize_model_id <- function(x) {
    gsub("[^A-Za-z0-9._-]", "_", x)
  }

  plot_benchmark_results <- function(results, plot_file = NULL) {
    if (!requireNamespace("ggplot2", quietly = TRUE)) {
      message("ggplot2 not available; skipping benchmark plot generation.")
      return(NULL)
    }

    plot_data <- results
    plot_data$model <- factor(
      plot_data$model,
      levels = plot_data$model[order(plot_data$elapsed_seconds, decreasing = TRUE)]
    )

    p <- ggplot2::ggplot(
      plot_data,
      ggplot2::aes(x = model, y = elapsed_seconds, fill = status)
    ) +
      ggplot2::geom_col(width = 0.7) +
      ggplot2::coord_flip() +
      ggplot2::labs(
        title = "patientChat benchmark by model",
        subtitle = "Task: 10 OMOP patients with diabetes + semaglutide",
        x = "Model",
        y = "Elapsed seconds",
        fill = "Status"
      ) +
      ggplot2::theme_minimal(base_size = 12)

    if (!is.null(plot_file)) {
      ggplot2::ggsave(
        filename = plot_file,
        plot = p,
        width = 10,
        height = 6,
        dpi = 120
      )
    }

    p
  }

  out_dir_env <- Sys.getenv("PATIENTGENERATOR_MODEL_BENCHMARK_DIR", unset = "")
  out_dir <- if (nzchar(out_dir_env)) {
    out_dir_env
  } else {
    file.path(tempdir(), paste0("patientChat_model_benchmark_", format(Sys.time(), "%Y%m%d_%H%M%S")))
  }
  dir.create(out_dir, recursive = TRUE, showWarnings = FALSE)

  results <- data.frame(
    model = character(),
    status = character(),
    elapsed_seconds = numeric(),
    person_count = integer(),
    output_file = character(),
    error = character(),
    stringsAsFactors = FALSE
  )

  for (model in models) {
    safe_model <- sanitize_model_id(model)
    output_name <- paste0("patient-chat-", safe_model)
    output_file <- file.path(out_dir, paste0(output_name, ".json"))

    started_at <- Sys.time()
    model_result <- tryCatch(
      {
        generator <- patientChat$new(model = model, echo = "none")
        generator$prompt(benchmark_prompt)
        generator$save(name = output_name, path = out_dir)

        cdm <- new_cdm()
        cdm$loadJsonTestSet(output_file)
        n_person <- nrow(cdm$person$data())

        status <- if (isTRUE(!is.na(n_person) && n_person == 10L)) "success" else "success_non_10"
        list(
          status = status,
          person_count = as.integer(n_person),
          error = NA_character_
        )
      },
      error = function(e) {
        list(
          status = "error",
          person_count = NA_integer_,
          error = conditionMessage(e)
        )
      }
    )
    elapsed <- as.numeric(difftime(Sys.time(), started_at, units = "secs"))

    results <- rbind(
      results,
      data.frame(
        model = model,
        status = model_result$status,
        elapsed_seconds = elapsed,
        person_count = model_result$person_count,
        output_file = output_file,
        error = model_result$error,
        stringsAsFactors = FALSE
      )
    )
  }

  results_file <- file.path(out_dir, "patientChat_model_benchmark_results.csv")
  utils::write.csv(results, results_file, row.names = FALSE)
  plot_file <- file.path(out_dir, "patientChat_model_benchmark_plot.png")
  benchmark_plot <- plot_benchmark_results(results, plot_file = plot_file)

  message("Model benchmark output directory: ", out_dir)
  message("Model benchmark results file: ", results_file)
  if (!is.null(benchmark_plot)) {
    message("Model benchmark plot file: ", plot_file)
    print(benchmark_plot)
  }
  print(results)

  # Keep assertions minimal: this test is intended to collect benchmark data.
  testthat::expect_equal(nrow(results), length(models))
  testthat::expect_true(file.exists(results_file))
  if (requireNamespace("ggplot2", quietly = TRUE)) {
    testthat::expect_true(file.exists(plot_file))
  }
})

test_that("Chaining LLM thought for synthetic patient generation", {
  skip_if_no_openai()
  
  # Instantiate patientGenerator
  expect_error({
    patientGenerator <- patientChat$new(model = "gpt")
  })
  expect_no_error({
    patientGenerator <- patientChat$new(model = "gpt-5.4-pro")
    patientGenerator <- patientChat$new(model = "gpt-5-nano")
    patientGenerator <- patientChat$new(model = "gpt-5-mini")
  })
  
  patientGenerator <- patientChat$new(model = "gpt-5.4")
  # patientGenerator$availableModels()
  
  # 1
  patientGenerator$prompt("5 female patients;
                          condition occurrence ovarian cancer with concept id 602306,
                          condition between 2015 and 2020.
                          All condition occurrences must end one year after index date")
  patientGenerator$save()
  
  expect_no_error({
    cdm <- TestGenerator::patientsCDM(testName = "patient-chat-test")
  })
  
  cdm$person |>
    pull(person_id) |>
    length() |>
    expect_equal(5)
  
  cdm$condition_occurrence |>
    pull(person_id) |>
    length() |>
    expect_equal(5)
  
  cdm$condition_occurrence |>
    pull(condition_concept_id) |>
    unique() |>
    expect_equal(602306)
  
  
  # 2
  patientGenerator$prompt("Eliminate 3 patients")
  patientGenerator$save()
  
  expect_no_error({
    cdm <- TestGenerator::patientsCDM(testName = "patient-chat-test")
  })
  
  cdm$person |>
    pull(person_id) |>
    length() |>
    expect_equal(2)
  
  cdm$condition_occurrence |>
    pull(person_id) |>
    length() |>
    expect_equal(2)
  
  cdm$condition_occurrence |>
    pull(condition_concept_id) |>
    unique() |>
    expect_equal(602306)
  
  
  # 3
  patientGenerator$prompt("Now fill out 10 more patients with ovarian cancer;
                          with 3 exposures of altretamine with concept_id code 1368823
                          to all patients in dataset;
                          Every exposure has an era of 30 days with a 6 day gap")
  patientGenerator$save()
  
  expect_no_error({
    cdm <- TestGenerator::patientsCDM(testName = "patient-chat-test")
  })
  
  cdm$person |>
    pull(person_id) |>
    length() |>
    expect_equal(12)
  
  cdm$condition_occurrence |>
    pull(person_id) |>
    length() |>
    expect_equal(12)
  
  cdm$drug_exposure |>
    pull(person_id) |>
    length() |>
    expect_equal(36)
  
  #4
  patientGenerator$prompt("Now, add 5 more patients to the previous dataset with ovarian cancer.
                          All patients must have also a measurement of stage 1 cancer with concept id code 1633306")
  patientGenerator$save()
  
  expect_no_error({
    cdm <- TestGenerator::patientsCDM(testName = "patient-chat-test")
  })
  
  cdm$person |>
    pull(person_id) |>
    length() |>
    expect_equal(17)
  
  cdm$condition_occurrence |>
    pull(condition_occurrence_id) |>
    length() |>
    expect_equal(17)
  
  cdm$measurement |>
    pull(person_id) |>
    length() |>
    expect_equal(17)
  
})


test_that("patientChat API flow creates loadable test set", {
  skip_if_no_openai()
  model <- pick_openai_model()

  expect_error(patientChat$new(model = "gpt"))

  generator <- patientChat$new(model = model, echo = "none")
  generator$prompt("Generate exactly 1 synthetic patient in OMOP-CDM v5.4 with one observation period.")

  tmp <- tempfile("pg_chat_")
  dir.create(tmp, recursive = TRUE)
  generator$save(name = "patient-chat-test", path = tmp)

  out_file <- file.path(tmp, "patient-chat-test.json")
  expect_true(file.exists(out_file))

  cdm <- new_cdm()
  expect_no_error(cdm$loadJsonTestSet(out_file))
  expect_gte(nrow(cdm$person$data()), 1)
})

test_that("patientChat with local codelist returns retrievable concepts", {
  codelist_path <- system.file("concept_sets", "ovarian_cancer_codelist.rds", package = "PatientGenerator")
  codelist_data <- readRDS(codelist_path)

  skip_if_no_openai()
  model <- pick_openai_model()

  generator <- patientChat$new(model = model, codelist_data = codelist_data, echo = "none")
  result <- generator$retrieveCodelist(concept_label = "ovarian", domain = "Condition")
  parsed <- jsonlite::fromJSON(result)

  expect_s3_class(parsed, "data.frame")
  expect_gt(nrow(parsed), 0)
})

skip_if_no_openai <- function() {
  testthat::skip_on_cran()
  key <- Sys.getenv("OPENAI_API_KEY", unset = "")
  if (!nzchar(key)) {
    testthat::skip("OPENAI_API_KEY not set")
  }
}

pick_openai_model <- function() {
  models <- tryCatch(
    PatientGenerator::availableModels(),
    error = function(e) testthat::skip(paste("Cannot list models:", conditionMessage(e)))
  )
  preferred <- c(
    "gpt-5.4",
    "gpt-5.2",
    "gpt-5",
    "gpt-4.1",
    "gpt-4o"
    )
  model <- preferred[preferred %in% models][1]
  if (is.na(model) || is.null(model)) {
    model <- models[1]
  }
  if (is.na(model) || !nzchar(model)) {
    testthat::skip("No available OpenAI model for this key")
  }
  model
}

test_that("patientChatNaive rejects non-JSON schema files", {
  bad_schema <- tempfile(fileext = ".txt")
  writeLines("not-json", bad_schema)

  expect_error(
    patientChatNaive(prompt = "Generate 1 patient", jsonSchemaPath = bad_schema),
    "jsonSchemaPath"
  )
})

test_that("patientChatNaive API output can be loaded by cdmConstructor", {
  skip_if_no_openai()
  model <- pick_openai_model()

  schema <- system.file("jsonSchemas", "cdm54schema-complete.json", package = "PatientGenerator")
  response <- patientChatNaive(
    prompt = "Generate exactly 1 patient in OMOP-CDM v5.4.",
    model = model,
    jsonSchemaPath = schema
  )

  out_json <- jsonlite::toJSON(
    response,
    dataframe = "rows",
    pretty = TRUE,
    null = "null",
    na = "null",
    auto_unbox = TRUE
  )

  out_file <- tempfile(fileext = ".json")
  write(out_json, out_file)

  cdm <- new_cdm()
  expect_no_error(cdm$loadJsonTestSet(out_file))
  expect_gte(nrow(cdm$person$data()), 1)
})

test_that("ellmer chat_structured output is loadable", {
  skip_if_no_openai()
  model <- pick_openai_model()

  schema <- system.file(
    "jsonSchemas",
    "cdm54schema-complete.json",
    package = "PatientGenerator"
    )
  chat <- ellmer::chat_openai(model = model)

  response <- chat$chat_structured(
    "Generate exactly one OMOP-CDM v5.4 patient with one condition occurrence.",
    type = ellmer::type_from_schema(path = schema)
  )

  out_json <- jsonlite::toJSON(
    response,
    dataframe = "rows",
    pretty = TRUE,
    null = "null",
    na = "null",
    auto_unbox = TRUE
  )

  out_file <- tempfile(fileext = ".json")
  write(out_json, out_file)

  cdm <- new_cdm()
  expect_no_error(cdm$loadJsonTestSet(out_file))
  expect_gte(nrow(cdm$person$data()), 1)
})

new_cdm <- function() {
  getFromNamespace("cdmConstructor", "PatientGenerator")$new()
}

new_cdm_table <- function(type) {
  getFromNamespace("cdmTable", "PatientGenerator")$new(type = type)
}

cdm_table_server <- getFromNamespace("cdmTableServer", "PatientGenerator")

personTable <- R6::R6Class("personTable",
                           inherit = cdmConstructor,
                           public = list(
                             data = NULL,
                             initialize = function() {
                               private$.columnNames <- columnNames("person")
                               private$.data <- private$.emptyTable()
                               self$data = private$.getData
                             },
                             add = function(...) {
                               person_id <- if (length(private$.data$person_id) == 0) {1L} else {private$.data$person_id[length(private$.data$person_id)] + 1L}
                               new_person_row <- private$.constructNewRow(person_id, ...)
                               private$.data <- rbindlist(list(private$.data, new_person_row))
                             },
                             update = function(person_id, ...) {
                               new_data <- list(...)
                               if (!all(names(new_data) %in% names(private$.columnNames))) {
                                 stop(glue::glue("Error: one column from c({glue::glue_collapse(names(new_data), sep = ', ')}) are not from the {private$.tableName} table"))
                               }
                               index_table <- which(private$.data[["person_id"]] == person_id)
                               if (length(index_table) > 0) {
                                 for (col_name in names(new_data)) {
                                   data.table::set(private$.data,
                                                   i = index_table,
                                                   j = col_name,
                                                   value = new_data[[col_name]])
                                 }
                               } else {
                                 warning(glue::glue("person_id not found in table"))
                               }
                             },
                             delete = function(person_id) {
                               index_table <- which(private$.data[["person_id"]] == person_id |> as.integer())
                               private$.data <- private$.data[-index_table]
                             },
                             reset = function() {
                               private$.data = private$.emptyTable()
                             }
                             ),
                           private = list(
                             .columnNames = NULL,
                             .data = NULL,
                             .constructNewRow = function(person_id, ...) {
                               checkmate::assertInteger(person_id)
                               new_data <- list(person_id = person_id, ...)
                               empty_row <- private$.emptyTable()
                               new_row <- rbindlist(list(empty_row, new_data), fill = TRUE)
                               return(new_row)
                             }
                             )
                           )

cdmTableUI <- function(id) {
  ns <- NS(id)
  card(
    full_screen = TRUE,
    card_body(
      fluidRow(
        column(
          2,
          actionButton(
            ns("add"),
            "Add"
          )
        ),
        # column(
        # 2,
        # actionButton(
        # ns("duplicate"),
        # "Duplicate"
        # )
        # ),
        column(
          2,
          conceptSearchUI(
            ns("concept_search"),
            buttonLabel = "Search Concept"
          )
        ),
        column(
          2,
          actionButton(
            ns("delete"),
            "Delete"
          )
        )
      ),
      br(),

      tagList(
        fluidRow(
          createInputs(
            ns,
            type = id,
            columns = columnNames(name = id, limit = 5) |> names()
          )
        )
      ),
      div(
        class = "reveal-on-full-screen",
        h5("Other columns"),
        fluidRow(
          tagList(
            createInputs(
              ns,
              type = id,
              columns = columnNames(name = id, limit = 5) |> names(),
              inverse = TRUE
            )
          )
        )
      ),
      fillable = FALSE
    )
  )
}

cdmTableServer <- function(
  id,
  cdm,
  person_id_selected,
  syncing
) {
  moduleServer(
    id,
    function(
      input,
      output,
      session
    ) {
      ns <- session$ns
      if (id == "condition_occurrence") {
        table_id <- "condition"
      } else {
        table_id <- id
      }

      table_event_id <- paste(
        id,
        "id",
        sep = "_"
      )
      if (id == "drug_exposure") {
        table_concept_id <- "drug_concept_id"
      } else {
        table_concept_id <- paste(
          table_id,
          "concept_id",
          sep = "_"
        )
      }
      table_start_date <- paste(
        table_id,
        "start_date",
        sep = "_"
      )
      table_end_date <- paste(
        table_id,
        "end_date",
        sep = "_"
      )
      columnList <- columnNames(
        name = id,
        limit = NULL
      ) |>
        names() |>
        tail(-2)

      ### ADD --------------------------------------------------------------------
      observeEvent(
        input$add,
        {
          # browser()
          # Require add button and the person id
          req(input$add)
          req(person_id_selected)
          # Create new event for that person in object
          cdm[[id]]$add(person_id = person_id_selected() |> as.integer())
          # Pull data from that person
          cdmTable <- cdm[[id]]$data() %>%
            dplyr::filter(
              person_id ==
                as.numeric(
                  person_id_selected()
                )
            )

          # Update interface to reflect addition
          updateSelectInput(
            session,
            "person_id",
            choices = cdmTable[["person_id"]] %>% unique(),
            selected = cdmTable[["person_id"]] %>% unique()
          )
          # When creating a new event the selected option is 
          # the last in the table
          updateSelectInput(
            session,
            table_event_id,
            choices = cdmTable[[table_event_id]],
            selected = cdmTable[[table_event_id]][length(
              cdmTable[[table_event_id]]
            )]
          )
        }
      )

      ### SEARCH CONCEPT ---------------------------------------------------------
      conceptSearchServer(
        id = "concept_search",
        onConceptSelected = function(conceptId) {
          shiny::updateTextInput(
            session,
            table_concept_id,
            value = as.character(conceptId)
          )
        },
        placeholderText = "e.g. Metformin"
      )

      # Observe the event id to update its corresponding fields in the interface
      observeEvent(
        input[[table_event_id]],
        {
          req(input[[table_event_id]])

          # Retrieve row
          cdmTable <- cdm[[id]]$data()
          cdmTableRow <- cdmTable |>
            filter(
              .data[[table_event_id]] == input[[table_event_id]] |> as.integer()
            )

          # Update
          updateInputs(
            session,
            ns,
            type = id,
            cdmTableRow = cdmTableRow,
            columns = columnList
          )
        },
        ignoreInit = TRUE
      )

      # UPDATE all other fields
      observe({
        inputs <- reactiveValuesToList(input)
        table_inputs <- inputs[names(inputs) %in% columnList]
        no_date_inputs <- table_inputs[grep(
          "date",
          names(table_inputs),
          invert = TRUE
        )]
        non_empty_inputs <- any(vapply(
          no_date_inputs,
          function(x) {
            is.character(x) && nzchar(x)
          },
          logical(1)
        ))
        req(non_empty_inputs)
        cdm_table_row <- cdm[[id]]$extractRow(
          event_id = input[[table_event_id]]
        )
        new_input_data <- as.data.frame(no_date_inputs)
        cdm_table_row[is.na(cdm_table_row)] <- ""
        new_input_data[is.na(new_input_data)] <- ""
        req(any(
          cdm_table_row[, names(no_date_inputs)] != new_input_data,
          na.rm = TRUE
        ))
        args <- c(
          list(event_id = as.integer(input[[table_event_id]])),
          no_date_inputs
        )
        # browser()
        do.call(cdm[[id]]$update, args)
      })

      # # Delete event
      observeEvent(input$delete, {
        req(input[[table_event_id]])

        # Delete function
        cdm[[id]]$delete(input[[table_event_id]])

        tableDeleted <- cdm[[id]]$data()

        if (tableDeleted[[table_event_id]] %>% length() == 0) {
          # browser()
          # Update to null all inputs...
          cdmTableRow <- tableDeleted
          updateInputs(
            session,
            ns,
            type = id,
            cdmTableRow = cdmTableRow,
            columns = columnNames(name = id, limit = 5) |> names()
          )
        } else {
          # ... if not, the value of the previous person remains after the person id is deleted
          choicesPerson <- tableDeleted[[table_event_id]][
            tableDeleted$person_id == person_id_selected()
          ]
          updateSelectInput(
            session,
            table_event_id,
            choices = choicesPerson,
            selected = as.numeric(input[[table_event_id]]) - 1
          )
        }
      })
      elongation_click <- reactiveVal(NULL)
      #
      # # Elongation
      observeEvent(
        list(input[[table_start_date]], input[[table_end_date]]),
        {
          print(syncing)
          req(!syncing())
          req(input[[table_start_date]])
          req(input[[table_end_date]])

          cdm[[id]]$updateDates(
            person_id = person_id_selected(),
            event_id = input[[table_event_id]],
            start_date = input[[table_start_date]],
            end_date = input[[table_end_date]]
          )
          elongation_click(Sys.time())
        },
        ignoreInit = TRUE
      )

      return(
        c(
          list(
            add_click = reactive(input$add),
            delete_click = reactive(input$delete),
            elongation_click = reactive(elongation_click())
          ),
          setNames(list(reactive(input[[table_event_id]])), table_event_id)
        )
      )
    }
  )
}

#' Useful to list available datasets in the testCases folder
#'
#' @param path Optional directory containing JSON test sets.
#' If NULL, the package resolves a default path with testthat integration.
#' @returns A list()
#' @export
#'
#' @importFrom stringr str_remove
getTestSets <- function(path = NULL) {

  testCasesDir <- testSetDir(path = path, create = TRUE)

  list.files(testCasesDir) |>
    stringr::str_remove(".json")

}

test_that("getTestSets supports explicit path", {
  tmp <- tempfile("pg_testsets_")
  dir.create(tmp, recursive = TRUE)
  write('{"person":[]}', file.path(tmp, "alpha.json"))
  write('{"person":[]}', file.path(tmp, "beta.json"))

  sets <- sort(getTestSets(path = tmp))
  expect_equal(sets, c("alpha", "beta"))
})

cdmTable <- R6::R6Class(
  "cdmTable",
  inherit = cdmConstructor,
  public = list(
  data = NULL,
  initialize = function(type) {
    private$.columnNames <- columnNames(type)
    private$.tableName <- type
    private$.data <- private$.emptyTable()
    self$data = private$.getData
  },
  reset = function() {
    private$.data = private$.emptyTable()
  },
  tableNameId = function() {
    private$.tableNameId()
  },
  tableNameConceptId = function() {
    private$.tableNameConceptId()
  },
  tableNameDate = function(name) {
    private$.tableNameDate(name)
  }
  ),
  private = list(
    .columnNames = NULL,
    .tableName = NULL,
    .data = NULL,
    .tableNameId = function() {
      name_id <- paste(
        private$.tableName,
        "id",
        sep = "_"
        )
      table_name_id <- private$.columnNames[
        ,
        .SD,
        .SDcols = grepl(
          name_id,
          names(private$.columnNames)
          )
        ] |>
        names()
      return(table_name_id)
    },
    .tableNameConceptId = function() {
      # browser()
      if (private$.tableName == "condition_occurrence") {
        table_id <- "condition"
      } else if (private$.tableName == "procedure_occurrence") {
        table_id <- "procedure"
        } else {
        table_id <- private$.tableName
      }
      if (private$.tableName == "drug_exposure") {
        table_concept_id <- "drug_concept_id"
      } else if (private$.tableName == "observation_period") {
        table_concept_id <- "period_type_concept_id"
      } else {
        table_concept_id <- paste(
          table_id,
          "concept_id",
          sep = "_"
        )
      }
      name_concept_id <- private$.columnNames[
        ,
        .SD,
        .SDcols = grepl(
          table_concept_id,
          names(private$.columnNames)
        )
      ] |>
        names()
      return(name_concept_id)
    },
    .tableNameDate = function(name) {
      # browser()
      checkmate::assertCharacter(name)
      if (!name %in% c("start", "end")) {
        stop("'name' should be 'start' or 'end'")
      }
      if (private$.tableName %in% c("condition_occurrence")) {
        table_name <- "condition"
      } else if (private$.tableName == "procedure_occurrence") {
        table_name <- "procedure"
      } else {
        table_name <- private$.tableName
      }
      name_date <- paste(
        table_name,
        name,
        "date",
        sep = "_"
        )
      if (name == "start" & private$.tableName == "measurement") {
        table_name_id <- "measurement_date"
      } else if (name == "start" & private$.tableName == "procedure_occurrence") {
        table_name_id <- "procedure_date"
      } else {
        table_name_id <- private$.columnNames[
          ,
          .SD,
          .SDcols = grepl(
            name_date,
            names(private$.columnNames)
            )
          ] |>
          names()
      }
      return(table_name_id)
    }
    )
  )

# Copyright 2023 DARWIN EU®
#
# This file is part of ReportGenerator
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
utils::globalVariables(c("cdmTableName",
                         "cdmFieldName",
                         "createCdmDeath",
                         "db",
                         ".SD",
                         "concept_name",
                         "domain_id",
                         "subject_id",
                         "cohort_start_date",
                         "cohort",
                         "cohort_end_date",
                         "delete_event",
                         "person_id",
                         "setNames",
                         "tail",
                         "updateTableDates"))

#' `patientDesigner()` is a visual interface based on D3 to construct test datasets for the OMOP-CDM
#'
#' @param path Optional folder containing JSON test sets.
#' If NULL, default path resolution keeps testthat integration.
#' @returns A Shiny app
#' @import r2d3 shiny bslib dplyr
#' @importFrom stats setNames
#' @importFrom utils tail
#' @importFrom data.table as.data.table set rbindlist
#' @export
patientDesigner <- function(path = NULL) {

  # bootswatch_themes <- c(
  #   "cerulean","cosmo","cyborg","darkly","flatly","journal","litera","lumen",
  #   "lux","materia","minty","morph","pulse","quartz","sandstone","simplex",
  #   "sketchy","slate","solar","spacelab","superhero","united","vapor","yeti","zephyr"
  # )

  ui <- page_fillable(
    tags$head(
      tags$style(HTML("
      /* Hide the element by default (when card is NOT full screen) */
      .bslib-card[data-full-screen='false'] .reveal-on-full-screen {
        display: none !important;
      }

      /* Optional: Add a transition or margin for smoother appearance */
      .reveal-on-full-screen {
        margin-top: 15px;
        padding-top: 15px;
        border-top: 1px solid #eee;
      }
    "))
    ),
    layout_sidebar(

    sidebar = sidebar(
        h4("PatientDesigner"),
        h6(actionLink(
            inputId = "new_test_set",
            label = strong("New Test Set"),
            icon = icon("pen-to-square"),
            class = "text-reset text-decoration-none"
          )
          ),
        br(),
        br(),
        h6(strong("Test Sets")),
        uiOutput("sidebar_file_list"),
        br(),
        br(),
        br(),
        br(),
        br(),
        br(),
        br(),
        br(),
        br(),
        br(),
        actionButton(
          "save_current",
          "Save Test Set",
          icon = icon("floppy-disk")
          ),
        downloadButton(
          "downloadTestSet",
          "Download Test Set",
          icon = icon("download")
          ),
        position = c("left"),
        open = "open"
        # selectInput("theme", "Bootswatch theme:", bootswatch_themes, selected = "flatly")
    ),
    layout_sidebar(
      sidebar = sidebar(
        h6(actionLink(
            inputId = "new_chat",
            label = strong("New Chat"),
            icon = icon("comment-dots"),
            class = "text-reset text-decoration-none"
            )
           ),
        position = "right", open = FALSE),
      tabsetPanel(
        tabPanel(
          "Person",
          personUI(id = "person"),
          tags$style(HTML("
  .well {
    padding: 1rem 1rem .15rem 1rem
  }
  "))
        )
      ),
      tabsetPanel(
        id = "cdm_table_tabs",
        tabPanel(
          "Observation Period",
          cdmTableUI(id = "observation_period"),
          value = "observation_period_module"
          ),
        tabPanel(
          "Condition Occurrence",
          cdmTableUI(id = "condition_occurrence"),
          value = "condition_occurrence_module"
          ),
        tabPanel(
          "Drug Exposure",
          cdmTableUI(id = "drug_exposure"),
          value = "drug_exposure_module"
          ),
        tabPanel(
          "Measurement",
          cdmTableUI(id = "measurement"),
          value = "measurement_module"
          ),
        tabPanel(
          "Procedure Occurrence",
          cdmTableUI(id = "procedure_occurrence"),
          value = "procedure_occurrence_module"
          )
      ),
      tabsetPanel(
        tabPanel(
          "Timeline",
          br(),
          d3Output(
            "d3",
            height = "1000px"
            )
        ),
        tabPanel(
          "Test Data",
          tableOutput("cdmData"),
          # verbatimTextOutput("cdmData"),
          tableOutput("personDataTable"),
          tableOutput("observationPeriodTable"),
          tableOutput("drugExposureTable"),
          tableOutput("conditionOccurrenceTable"),
          tableOutput("measurementTable"),
          tableOutput("procedureOccurrenceTable")
          )
      ),
      border = FALSE
    ),
    border_radius = FALSE,
    fillable = TRUE,
    class = "p-0"
  ),
  padding = c(0),
  title = "OHDSI - PatientGenerator - PatientDesigner",
  theme = bs_theme(version = 5, bootswatch = "zephyr")  # initial theme
  )



  server <- function(input, output, session) {

    # Swap theme in real time
    # observeEvent(input$theme, ignoreInit = TRUE, {
    session$setCurrentTheme(
      bs_theme(
        version = 5,
        bootswatch = "zephyr"
        )
      )
    # })

    # TRIGGERS
    file_refresh_trigger <- reactiveVal(0)
    loaded_listeners <- reactiveVal(character(0))
    data_version <- reactiveVal(0)

    # TestCases folder
    get_test_dir <- function() {
      testSetDir(path = path, create = TRUE)
    }

    # Create CDM object
    cdm <- cdmConstructor$new()

    # Wipe clean
    observeEvent(input$new_test_set, {
      # browser()
      cdm$reset()
      data_version(data_version() + 1)
    })

    ##### Update saved file in sidebar
    current_files <- reactive({
      file_refresh_trigger()
      list.files(
        get_test_dir(),
        pattern = "\\.json$",
        full.names = FALSE
        )
    })

    # Render list UI
    output$sidebar_file_list <- renderUI({
      files <- current_files()

      tagList(
        lapply(files, function(f) {
          # ID includes extension to be unique and consistent
          # Display name removes extension for looks
          actionLink(
            inputId = paste0("link_", f),
            label = tools::file_path_sans_ext(f),
            class = "text-reset text-decoration-none d-block mb-1"
          )
        })
      )
    })

    # Handles old and new files
    observe({
      files <- current_files()
      existing <- loaded_listeners()
      new_files <- setdiff(files, existing)

      lapply(new_files, function(filename) {

        id <- paste0("link_", filename)

        observeEvent(input[[id]], {
          path <- file.path(get_test_dir(), filename)
          cdm$loadJsonTestSet(path)
          data_version(data_version() + 1)
        })
      })

      if (length(new_files) > 0) {
        loaded_listeners(c(existing, new_files))
      }
      })

    observeEvent(input$save_current, {
      showModal(modalDialog(
        title = "Save Test Set",
        textInput(
          "new_filename",
          "Filename (no extension):",
          placeholder = "my_test"
          ),
        footer = tagList(
          modalButton("Cancel"),
          actionButton(
            "confirm_save",
            "Save",
            class = "btn-primary"
            )
        )
      ))
    })

    observeEvent(input$confirm_save, {
      req(input$new_filename)

      new_name <- paste0(
        tools::file_path_sans_ext(input$new_filename),
        ".json"
        )
      path <- file.path(
        get_test_dir(),
        new_name
        )
      write(
        cdm$getCdmData(),
        path
        )

      file_refresh_trigger(file_refresh_trigger() + 1)
      removeModal()
    })


    ##### Load JSON Test Set
    lapply(
      getTestSets(
        path = get_test_dir()),
      function(filename) {
        id <- paste0("link_", filename)
        observeEvent(input[[id]], {
          path <- file.path(
            get_test_dir(),
            paste(
              filename,
              "json",
              sep = "."
              )
            )
        cdm$loadJsonTestSet(path)
        data_version(data_version() + 1)
      })
    })

    ##### PERSON TABLE

    # Person server module - Create, delete and update
    person_module <- personServer(
      id = "person",
      cdm = cdm,
      trigger = data_version
      )

    # Render person table
    output$personDataTable <- renderTable({
      data_version()
      person_module()
      cdm$person$data()
    })

    # After person selection
    # Filters and updates observation/drug exposure fields
    observeEvent(person_module(), {
      req(person_module())

      updateTableIdsNs(
        cdm = cdm,
        type = "observation_period",
        input_person_id = person_module,
        session = session
        )
      updateTableIdsNs(
        cdm = cdm,
        type = "condition_occurrence",
        input_person_id = person_module,
        session = session
        )
      updateTableIdsNs(
        cdm = cdm,
        type = "drug_exposure",
        input_person_id = person_module,
        session = session
        )

    }, ignoreInit = TRUE)

    ##### OBSERVATION PERIOD
    # - Each section shows only data from the selected individual
    #   in the person section

    # Module - Create, delete and update
    observation_period_module <- cdmTableServer(
      id = "observation_period",
      cdm = cdm,
      person_id_selected = person_module,
      syncing = syncing
      )

    # Render observation period table
    output$observationPeriodTable <- renderTable({
      data_version()
      observation_period_module$add_click()
      observation_period_module$delete_click()
      observation_period_module$elongation_click()
      cdm$observation_period$data()
    })

    ##### DRUG EXPOSURE TABLE

    drug_exposure_module <- cdmTableServer(
      id = "drug_exposure",
      cdm = cdm,
      person_id_selected = person_module,
      syncing = syncing
      )


    # Render drug exposure table
    output$drugExposureTable <- renderTable({
      data_version()
      drug_exposure_module$add_click()
      drug_exposure_module$delete_click()
      drug_exposure_module$elongation_click()
      cdm$drug_exposure$data()
    })

    # CONDITION OCCURRENCE TABLE
    condition_occurrence_module <- cdmTableServer(
      id = "condition_occurrence",
      cdm = cdm,
      person_id_selected = person_module,
      syncing = syncing
      )

    # Render drug exposure table
    output$conditionOccurrenceTable <- renderTable({
      data_version()
      condition_occurrence_module$add_click()
      condition_occurrence_module$delete_click()
      condition_occurrence_module$elongation_click()
      cdm$condition_occurrence$data()
    })

    # MEASUREMENT TABLE
    measurement_module <- cdmTableServer(
      id = "measurement",
      cdm = cdm,
      person_id_selected = person_module,
      syncing = syncing
      )

    output$measurementTable <- renderTable({
      data_version()
      measurement_module$add_click()
      measurement_module$delete_click()
      measurement_module$elongation_click()
      cdm$measurement$data()
    })
    
    # PROCEDURE OCCURRENCE TABLE
    procedure_occurrence_module <- cdmTableServer(
      id = "procedure_occurrence",
      cdm = cdm,
      person_id_selected = person_module,
      syncing = syncing
    )
    
    output$procedureOccurrenceTable <- renderTable({
      data_version()
      procedure_occurrence_module$add_click()
      procedure_occurrence_module$delete_click()
      procedure_occurrence_module$elongation_click()
      cdm$procedure_occurrence$data()
    })

    # CDM Data Timeline
    cdmDataTimeline <- reactive({
      pid <- suppressWarnings(as.numeric(person_module()))
      req(!is.na(pid), length(pid) == 1)
      # browser()
      cdm$getCdmDataTimeline() %>%
        dplyr::filter(.data$person_id == pid)
    }) %>% bindEvent(
      data_version(),
      person_module(),
      observation_period_module$add_click(),
      observation_period_module$delete_click(),
      observation_period_module$elongation_click(),
      drug_exposure_module$add_click(),
      drug_exposure_module$delete_click(),
      drug_exposure_module$elongation_click(),
      condition_occurrence_module$add_click(),
      condition_occurrence_module$delete_click(),
      condition_occurrence_module$elongation_click(),
      measurement_module$add_click(),
      measurement_module$delete_click(),
      measurement_module$elongation_click(),
      procedure_occurrence_module$add_click(),
      procedure_occurrence_module$delete_click(),
      procedure_occurrence_module$elongation_click(),
      ignoreInit = FALSE
      )

    # Render cdm table
    output$cdmData <- renderTable({
      req(cdmDataTimeline)
      cdmDataTimeline()
    })

    ## UPDATE DATA FROM D3

    # Drag behavior - start, move and end

    syncing <- reactiveVal(FALSE)

    observeEvent(input$bar_start, {
      # browser()
      syncing(TRUE)
      start_data <- input$bar_start
      type_module <- paste(
        start_data$type,
        "module",
        sep = "_"
        )
      updateTabsetPanel(
        session,
        "cdm_table_tabs",
        selected = type_module
        )
      updateTablePersonEventIdsNs(
        cdm,
        type = start_data$type,
        input_person_id = start_data$person_id,
        input_event_id = start_data$event_id,
        session
        )
    })

    observeEvent(input$bar_end, {
      update_data <- input$bar_end
      person_id <- update_data$person_id
      event_id <- update_data$event_id
      type <- update_data$type
      print("END DATA:")
      update_data$start_date %>% print()
      update_data$end_date %>% print()
      cdm[[type]]$updateDates(
        person_id = person_id,
        event_id = event_id,
        start_date = update_data$start_date,
        end_date = update_data$end_date
        )
      updateTableDatesNs(
        cdm = cdm,
        type = type,
        input_person_id = person_id,
        input_event_id = event_id,
        start_date = update_data$start_date,
        end_date = update_data$end_date,
        session = session,
        input = input,
        syncing = syncing
        )
      syncing(FALSE)

    })

    output$downloadTestSet <- downloadHandler(
      filename = function() {
        paste("patientDesigner", ".json", sep = "")
      },
      content = function(file) {
        write(cdm$getCdmData(), file)
      }
    )

    ##### D3 TIMELINE
    output$d3 <- renderD3({
      r2d3(
        data = cdmDataTimeline(),
        script = system.file("d3/cdm_timeline.js", package = "PatientGenerator"),
        height = 1000
      )
    })
  }
  shinyApp(ui = ui, server = server)
}

createInputs <- function(ns, type, columns, inverse = FALSE) {
  if (inverse) {
    columns <- setdiff(columnNames(type) |> names(), columns)
  }
  Map(function(col_name) {
    if (stringr::str_detect(col_name, "date")) {
      column(2,
             dateInput(ns(col_name),
                       label = stringr::str_to_sentence(stringr::str_replace_all(col_name, "_", " ")))
      )
    } else if (stringr::str_detect(col_name, "concept") & !stringr::str_detect(col_name, "gender")) {
      column(2,
             textInput(ns(col_name),
                       label = stringr::str_to_sentence(stringr::str_replace_all(col_name, "_", " ")))
      )
    } else {
      if (!stringr::str_detect(col_name, "person_id")) {
        choices <- choicesList(tableName = type)
      } else {
        choices <- NULL
      }
        column(2,
               selectizeInput(ns(col_name),
                              label = stringr::str_to_sentence(stringr::str_replace_all(col_name, "_", " ")),
                              choices = choices[[col_name]],
                              selected = NULL,
                              options = list(dropdownParent = "body"))
        )
    }
  }, columns)
}

updateInputs <- function(session, ns, type, cdmTableRow, columns) {


  Map(function(col_name) {
    # browser()
    if (stringr::str_detect(col_name, "date")) {
      updateDateInput(
        session = session,
        inputId = col_name,
        value = cdmTableRow[[col_name]],
        min = NULL,
        max = NULL
       )
    } else if (stringr::str_detect(col_name, "concept") & !stringr::str_detect(col_name, "gender")) {
      updateTextInput(
        session = session,
        inputId = col_name,
        value = cdmTableRow[[col_name]]
        )
    } else {
      updateSelectizeInput(
        session = session,
        inputId = col_name,
        selected = cdmTableRow[[col_name]],
        options = list(dropdownParent = "body")
        )
    }
  }, columns)
}

listInputParameters <- function(ns, input, type, columns) {
  lapply(columns, function(col) {
    input[[col]]
  })
}

updatePersonInputs <- function(cdmPersonData,
                               input_person_id,
                               ids,
                               session) {
  personTableSelection <- cdmPersonData %>%
    filter(person_id == input_person_id)
  Map(function(id) {
    updateSelectInput(session,
                      id,
                      selected = personTableSelection[[id]])
  }, ids)
}

updateTableIds <- function(cdm, type = "drug_exposure", input_person_id, session) {

  # Access names
  table_name <- glue::glue("get_{type}_table")
  table_person_id <- glue::glue("{type}_person_id")
  table_event_id <- glue::glue("{type}_id")
  table_module <- glue::glue("{type}_module")


  # Get observation period data and filter for the selected person_id
  cdmTable <- cdm[[table_name]]() %>%
    filter(person_id == as.numeric(input_person_id))

  # Update observation period fields
  updateSelectInput(session, table_person_id,
                    choices = cdmTable[["person_id"]] %>% unique(),
                    selected = cdmTable[["person_id"]] %>% unique())
  updateSelectInput(session, table_event_id,
                    choices = cdmTable[[table_event_id]],
                    selected = cdmTable[[table_event_id]][length(cdmTable[[table_event_id]])])

}

updateTableIdsNs <- function(cdm, type = "drug_exposure", input_person_id, session) {
  # browser()
  # Access names
  table_person_id <- glue::glue("{type}_person_id")
  table_event_id <- glue::glue("{type}_id")

  # Get observation period data and filter for the selected person_id
  cdmTable <- cdm[[type]]$data() %>%
    filter(person_id == as.numeric(input_person_id()))

  ns_module <- NS(type)

  # Update observation period fields
  updateSelectInput(session, ns_module(table_person_id),
                    choices = cdmTable[["person_id"]] %>% unique(),
                    selected = cdmTable[["person_id"]] %>% unique())
  updateSelectInput(session, ns_module(table_event_id),
                    choices = cdmTable[[table_event_id]],
                    selected = cdmTable[[table_event_id]][length(cdmTable[[table_event_id]])])

}

updateTablePersonEventIdsNs <- function(cdm, type = "drug_exposure", input_person_id, input_event_id, session) {
  # browser()
  # Access names
  table_person_id <- glue::glue("{type}_person_id")
  table_event_id <- glue::glue("{type}_id")

  p_id <- if (is.function(input_person_id)) input_person_id() else input_person_id
  req(p_id)

  e_id <- if (is.function(input_event_id)) input_event_id() else input_event_id
  req(e_id)

  # Get observation period data and filter for the selected person_id
  cdmTable <- cdm[[type]]$data() %>%
    filter(person_id == as.numeric(p_id))

  ns_module <- NS(type)

  # Update observation period fields
  updateSelectInput(session, ns_module(table_person_id),
                    choices = cdmTable[["person_id"]] %>% unique(),
                    selected = cdmTable[["person_id"]] %>% unique())
  updateSelectInput(session, ns_module(table_event_id),
                    choices = cdmTable[[table_event_id]],
                    selected = e_id)

}

updateTableDatesNs <- function(cdm,
                               type = "drug_exposure",
                               input_person_id,
                               input_event_id,
                               start_date,
                               end_date,
                               session,
                               input,
                               syncing) {

  # browser()
  if (type == "condition_occurrence") {
    type_corrected <- "condition"
  } else {
    type_corrected <- type
  }

  # Access names
  table_start_date <- glue::glue("{type_corrected}_start_date")
  table_end_date <- glue::glue("{type_corrected}_end_date")
  table_id <- glue::glue("{type}_id")

  p_id <- if (is.function(input_person_id)) { input_person_id() } else { input_person_id }
  req(p_id)

  e_id <- if (is.function(input_event_id)) input_event_id() else input_event_id
  req(e_id)

  # Get observation period data and filter for the selected person_id
  # cdmTable <- cdm[[type]]$data() %>%
  #   filter(person_id == as.numeric(p_id)) %>%
  #   filter(.data[[table_id]] == as.numeric(e_id))

  ns_module <- NS(type)
  # Update observation period fields
  freezeReactiveValue(input, ns_module(table_start_date))
  updateDateInput(session, ns_module(table_start_date), value = start_date)

  freezeReactiveValue(input, ns_module(table_end_date))
  updateDateInput(session, ns_module(table_end_date), value = end_date)

}

#' Concept search Shiny module
#'
#' Reusable UI and server for searching the vocabulary (Hecate) and selecting
#' a concept. Shows a button that opens a modal with text input, search button,
#' and results in a DT table. Optional callback when a concept is selected.
#'
#' @param id Module namespace id.
#' @param buttonLabel Label for the trigger button (default \code{"Concept search"}).
#' @param onConceptSelected Optional function of one argument \code{conceptId}
#'   called when user selects a concept; modal is closed after.
#' @param placeholderText Character string used as placeholder text in the
#'   search input field.
#' @name conceptSearchModule
NULL

#' @describeIn conceptSearchModule UI for the concept search: a single button that opens the search modal.
#' @export
conceptSearchUI <- function(id, buttonLabel = "Concept search") {
  ns <- shiny::NS(id)
  shiny::actionButton(ns("open_concept_search"), buttonLabel)
}

#' @describeIn conceptSearchModule Server for the concept search modal (text input, search, DT, close).
#' @export
conceptSearchServer <- function(id, onConceptSelected = NULL, placeholderText = "") {
  shiny::moduleServer(id, function(input, output, session) {
    ns <- session$ns

    searchResults <- shiny::reactiveVal(NULL)
    standardOnlyFilter <- shiny::reactiveVal(TRUE)

    shiny::observeEvent(input$open_concept_search, {
      searchResults(NULL)
      standardOnlyFilter(TRUE)
      shiny::showModal(
        shiny::modalDialog(
          title = "Search vocabulary",
          shiny::tags$style(shiny::HTML(".modal-dialog { max-width: 90%; width: 90%; }")),
          shiny::tags$script(shiny::HTML(paste0(
            "$(document).on('keydown', '#", ns("concept_query"), "', function(e) {",
            "  if (e.which === 13) { e.preventDefault(); $('#", ns("search_vocabulary"), "').click(); }",
            "});",
            "$(document).on('keydown', function(e) {",
            "  if (e.which === 27 && $('.modal.show').length) {",
            "    Shiny.setInputValue('", ns("modal_escape"), "', Math.random(), {priority: 'event'});",
            "  }",
            "});"
          ))),
          shiny::fluidRow(
            shiny::column(2,
              shiny::textInput(ns("concept_query"), "Query", placeholder = placeholderText)
            ),
            shiny::column(1, style = "padding-top: 32px;",
              shiny::actionButton(ns("search_vocabulary"), "Search", class = "btn-primary")
            ),
            shiny::column(4, style = "padding-top: 40px;",
              shiny::checkboxInput(ns("standard_concepts_only"), "Standard Concepts only", value = TRUE)
            )
          ),
          shiny::hr(),
          DT::DTOutput(ns("concept_results_table")),
          footer = shiny::tagList(
            shiny::modalButton("Close")
          ),
          size = "l"
        )
      )
    })

    shiny::observeEvent(input$search_vocabulary, {
      query <- shiny::req(input$concept_query)
      if (nchar(trimws(query)) < 1) return()
      df <- hecateSearch(query, limit = 50)
      if (!is.null(df) && nrow(df) > 0 && "searchScore" %in% names(df)) {
        df <- df[order(-df$searchScore), , drop = FALSE]
      }
      searchResults(df)
    })

    shiny::observeEvent(input$standard_concepts_only, {
      standardOnlyFilter(isTRUE(input$standard_concepts_only))
    }, ignoreNULL = TRUE)

    output$concept_results_table <- DT::renderDT({
      df <- searchResults()
      if (is.null(df) || nrow(df) == 0) {
        return(NULL)
      }
      if (isTRUE(standardOnlyFilter()) && "standardConcept" %in% names(df)) {
        df <- df[!is.na(df$standardConcept) & as.character(df$standardConcept) == "S", , drop = FALSE]
      }
      if (nrow(df) == 0) {
        return(NULL)
      }
      # Add a Select column with links (build onclick inline to avoid htmltools::JS dependency)
      inputId <- ns("selected_concept_row")
      df$Select <- vapply(
        seq_len(nrow(df)),
        function(i) {
          rowIdx <- i - 1L
          paste0(
            "<a href=\"#\" onclick=\"Shiny.setInputValue('", inputId, "', ", rowIdx,
            ", {priority: 'event'}); return false;\">Select</a>"
          )
        },
        character(1)
      )
      # Show key columns including standardConcept; keep Select last
      cols <- c("conceptId", "conceptName", "standardConcept", "domainId", "vocabularyId", "conceptCode", "Select")
      cols <- intersect(cols, names(df))
      df <- df[, cols, drop = FALSE]
      DT::datatable(
        df,
        escape = FALSE,
        selection = "none",
        options = list(
          pageLength = 10,
          dom = "frtip",
          search = list(smart = TRUE)
        )
      )
    })

    shiny::observeEvent(input$selected_concept_row, {
      row <- as.integer(shiny::req(input$selected_concept_row)) + 1L
      df <- searchResults()
      if (is.null(df)) return()
      if (isTRUE(standardOnlyFilter()) && "standardConcept" %in% names(df)) {
        df <- df[!is.na(df$standardConcept) & as.character(df$standardConcept) == "S", , drop = FALSE]
      }
      if (row < 1 || row > nrow(df)) return()
      conceptId <- df$conceptId[row]
      if (!is.null(onConceptSelected)) {
        onConceptSelected(conceptId)
      }
      shiny::removeModal()
    })

    shiny::observeEvent(input$modal_escape, {
      shiny::removeModal()
    })
  })
}

#' Internal helper to locate where patient test sets are stored
#' @noRd
testSetDir <- function(path = NULL, create = TRUE) {
  if (is.null(path)) {
    path <- tryCatch(
      testthat::test_path("testCases"),
      error = function(e) NULL
    )
  }

  if (is.null(path)) {
    local_testthat_dir <- file.path(getwd(), "tests", "testthat", "testCases")
    if (dir.exists(dirname(local_testthat_dir))) {
      path <- local_testthat_dir
    }
  }

  if (is.null(path)) {
    path <- getOption(
      "PatientGenerator.testSetDir",
      getOption("patientGenerator.testSetDir", NULL)
    )
  }

  if (is.null(path)) {
    path <- file.path(tools::R_user_dir("PatientGenerator", which = "data"), "testCases")
  }

  if (isTRUE(create) && !dir.exists(path)) {
    dir.create(path, recursive = TRUE, showWarnings = FALSE)
  }

  path
}

test_that("retrieveCodelist helper and tool object are available", {
  retrieve_fn <- getFromNamespace("retrieveCodelist", "PatientGenerator")
  tool_obj <- getFromNamespace("retrieveCodelistTool", "PatientGenerator")

  result <- retrieve_fn(concept_label = "ovarian", domain = "Condition")
  parsed <- jsonlite::fromJSON(result)

  expect_s3_class(parsed, "data.frame")
  expect_gt(nrow(parsed), 0)
  expect_false(is.null(tool_obj))
})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/patientChat_naive.R
\name{availableModels}
\alias{availableModels}
\title{\code{availableModels()} If the API key is valid in the system, returns available models to the user from the LLM provider.}
\usage{
availableModels()
}
\value{
A string list with the id of a vailable models.
}
\description{
OpenAI is the only one currently supported.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilsPatientDesigner.R
\name{getTestSets}
\alias{getTestSets}
\title{Useful to list available datasets in the testCases folder}
\usage{
getTestSets(path = NULL)
}
\arguments{
\item{path}{Optional directory containing JSON test sets.
If NULL, the package resolves a default path with testthat integration.}
}
\value{
A list()
}
\description{
Useful to list available datasets in the testCases folder
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/patientsDesigner.R
\name{patientDesigner}
\alias{patientDesigner}
\title{\code{patientDesigner()} is a visual interface based on D3 to construct test datasets for the OMOP-CDM}
\usage{
patientDesigner(path = NULL)
}
\arguments{
\item{path}{Optional folder containing JSON test sets.
If NULL, default path resolution keeps testthat integration.}
}
\value{
A Shiny app
}
\description{
\code{patientDesigner()} is a visual interface based on D3 to construct test datasets for the OMOP-CDM
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hecateSearch.R
\name{hecateSearch}
\alias{hecateSearch}
\title{Search Hecate concepts and return results as a data frame}
\usage{
hecateSearch(
  query,
  vocabularyId = NULL,
  standardConcept = NULL,
  domainId = NULL,
  conceptClassId = NULL,
  limit = 20,
  client = hecateClient()
)
}
\arguments{
\item{query}{Character(1); search query (required).}

\item{vocabularyId}{Character(1) or NULL; optional vocabulary filter (comma-separated).}

\item{standardConcept}{Character(1) or NULL; e.g. \code{"S"} (Standard), \code{"C"} (Classification).}

\item{domainId}{Character(1) or NULL; optional domain filter (comma-separated).}

\item{conceptClassId}{Character(1) or NULL; optional concept class filter.}

\item{limit}{Integer(1); max results (default 20, max 150).}

\item{client}{Hecate client (default \code{hecateClient()}).}
}
\value{
Data frame of search results, or NULL if an error occurred (API error or bad response shape).
}
\description{
Search Hecate concepts and return results as a data frame
}
\examples{
\dontrun{
# Simple search
df <- hecateSearch("diabetes")

# Search with filters
df <- hecateSearch("hypertension", domainId = "Condition", limit = 10)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/conceptSearchModule.R
\name{conceptSearchModule}
\alias{conceptSearchModule}
\alias{conceptSearchUI}
\alias{conceptSearchServer}
\title{Concept search Shiny module}
\usage{
conceptSearchUI(id, buttonLabel = "Concept search")

conceptSearchServer(id, onConceptSelected = NULL, placeholderText = "")
}
\arguments{
\item{id}{Module namespace id.}

\item{buttonLabel}{Label for the trigger button (default \code{"Concept search"}).}

\item{onConceptSelected}{Optional function of one argument \code{conceptId}
called when user selects a concept; modal is closed after.}

\item{placeholderText}{Character string used as placeholder text in the
search input field.}
}
\description{
Reusable UI and server for searching the vocabulary (Hecate) and selecting
a concept. Shows a button that opens a modal with text input, search button,
and results in a DT table. Optional callback when a concept is selected.
}
\section{Functions}{
\itemize{
\item \code{conceptSearchUI()}: UI for the concept search: a single button that opens the search modal.

\item \code{conceptSearchServer()}: Server for the concept search modal (text input, search, DT, close).

}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/patientChat_naive.R
\name{patientChatNaive}
\alias{patientChatNaive}
\title{\code{patientChatNaive()} is a grapper for the ellmer package to send prompts and send the output test set to an LLM. Requires a valid API key.}
\usage{
patientChatNaive(
  prompt = "### Give me a sample of five patients",
  model = "gpt-5.2",
  jsonSchemaPath = NULL
)
}
\arguments{
\item{prompt}{A prompt to the LLM, in character or JSON response.}

\item{model}{The model used by the LLM. Currently only OpenAI models are accepted.}

\item{jsonSchemaPath}{Path to a JSON schema used to structure the response.}
}
\value{
A JSON response that includes: the natural language answer from the LLM and a JSON with test set patients in accordance to the provided schema.
}
\description{
Priorities:
- Accepts a prompt as an input.
- Produces a test set in accordance to the provided JSON schema.
- Utilizes tools such as CodelistGenerator or Hecate to look up for functions.
- Accepts a subsequent prompt with a test set that the LLM has to use as a context

One function for this tasks allow us to:
- Test the test sets created by the LLM.
- Test prompt engineering.
- Test integration of tools functionality.
- Allow us to create fast a small set of patients to test analytical packages.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hecateSearch.R
\name{hecateClient}
\alias{hecateClient}
\title{Create a Hecate API client}
\usage{
hecateClient(baseUrl = NULL, timeoutMs = NULL, apiKey = NULL)
}
\arguments{
\item{baseUrl}{Base URL of the Hecate API (default from config).}

\item{timeoutMs}{Timeout in milliseconds (default from config).}

\item{apiKey}{Optional API key for authorization (default from config).}
}
\value{
A client object with class \code{hecate_client}.
}
\description{
Create a Hecate API client
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hecateSearch.R
\name{nullOr}
\alias{nullOr}
\alias{\%||\%}
\title{Null coalescing operator}
\usage{
x \%||\% y
}
\arguments{
\item{x}{First value (any type).}

\item{y}{Fallback value when \code{x} is NULL.}
}
\value{
\code{x} if not NULL, otherwise \code{y}.
}
\description{
Null coalescing operator
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/patientChat.R
\name{patientChat}
\alias{patientChat}
\title{\code{patientChat()} generates synthetic patients in the OMOP-CDM using an LLM API.}
\value{
A JSON response that includes: the natural language answer from the LLM and
a JSON with test set patients in accordance to the provided schema.
}
\description{
Requires an OPEN_AI_KEY in ~/.Renviron.
After that just sent a prompt and save() the results.
The JSON file can be used as an OMOP-CDM patient test set.
}
\details{
Accepts a prompt as input. Produces a test set using a
structured JSON schema. Utilizes tools such as CodelistGenerator or Hecate
to look up concept IDs. Accepts subsequent prompts to modify existing test
sets that the LLM uses as context.

This class allows testing patient sets created by the LLM, prompt engineering,
integration of search tools and functionality, and creating a set of patients
to test analytical packages.
}
\examples{
\dontrun{
generator <- patientChat$new()
generator$prompt("Give me 5 patients")
generator$save("my_test")
}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{chat}}{An ellmer chat instance}

\item{\code{json_schema_path}}{JSON schema to output structured results}

\item{\code{response}}{Ouput from the LLM}

\item{\code{codelist}}{A codelist with details to search for concepts ids}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-patientChat-new}{\code{patientChat$new()}}
\item \href{#method-patientChat-prompt}{\code{patientChat$prompt()}}
\item \href{#method-patientChat-json_response}{\code{patientChat$json_response()}}
\item \href{#method-patientChat-output}{\code{patientChat$output()}}
\item \href{#method-patientChat-retrieveCodelist}{\code{patientChat$retrieveCodelist()}}
\item \href{#method-patientChat-save}{\code{patientChat$save()}}
\item \href{#method-patientChat-availableModels}{\code{patientChat$availableModels()}}
\item \href{#method-patientChat-clone}{\code{patientChat$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-patientChat-new"></a>}}
\if{latex}{\out{\hypertarget{method-patientChat-new}{}}}
\subsection{Method \code{new()}}{
Create a new chat to create JSON test sets for OMOP-CDM.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{patientChat$new(
  system_prompt = NULL,
  model = "gpt-5.4",
  jsonSchemaPath = NULL,
  echo = c("none", "output", "all"),
  codelist_data = NULL
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{system_prompt}}{Initial system prompt to impose behaviour to the LLM}

\item{\code{model}}{Such as "gpt-5.3". For a complete list, call patientChat$availableModels()}

\item{\code{jsonSchemaPath}}{The JSON schema to structure output from LLM}

\item{\code{echo}}{How the output will be displayed in the console}

\item{\code{codelist_data}}{A codelist with details to search for concepts ids}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new \code{Person} object.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-patientChat-prompt"></a>}}
\if{latex}{\out{\hypertarget{method-patientChat-prompt}{}}}
\subsection{Method \code{prompt()}}{
Prompt to request data from LLM API
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{patientChat$prompt(prompt)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{prompt}}{A query in character.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-patientChat-json_response"></a>}}
\if{latex}{\out{\hypertarget{method-patientChat-json_response}{}}}
\subsection{Method \code{json_response()}}{
Output in JSON format
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{patientChat$json_response()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-patientChat-output"></a>}}
\if{latex}{\out{\hypertarget{method-patientChat-output}{}}}
\subsection{Method \code{output()}}{
Returns the chat object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{patientChat$output()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-patientChat-retrieveCodelist"></a>}}
\if{latex}{\out{\hypertarget{method-patientChat-retrieveCodelist}{}}}
\subsection{Method \code{retrieveCodelist()}}{
Retrieves and filters data from codelist_data
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{patientChat$retrieveCodelist(concept_label = "Stage 1", domain = "Measurement")}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{concept_label}}{Filters the concept_name in the codelist with details}

\item{\code{domain}}{Filters the domain in the codelist with details.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-patientChat-save"></a>}}
\if{latex}{\out{\hypertarget{method-patientChat-save}{}}}
\subsection{Method \code{save()}}{
Saves the JSON test set to disk.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{patientChat$save(name = "patient-chat-test", path = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{name}}{Name of the file}

\item{\code{path}}{To save the file.
If NULL, the package first tries \code{testthat::test_path("testCases")},
then checks \code{options(PatientGenerator.testSetDir = "...")}, and finally
falls back to the package user data directory.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-patientChat-availableModels"></a>}}
\if{latex}{\out{\hypertarget{method-patientChat-availableModels}{}}}
\subsection{Method \code{availableModels()}}{
Retrieves available models from the LLM API.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{patientChat$availableModels()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-patientChat-clone"></a>}}
\if{latex}{\out{\hypertarget{method-patientChat-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{patientChat$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}
